from unittest import TestCase

from piccolo.columns.column_types import Bytea
from piccolo.table import Table


class MyTable(Table):
    token = Bytea()


class MyTableDefault(Table):
    """
    Test the different default types.
    """

    token = Bytea()
    token_bytes = Bytea(default=b"my-token")
    token_bytearray = Bytea(default=bytearray(b"my-token"))
    token_none = Bytea(default=None, null=True)


class TestBytea(TestCase):
    def setUp(self):
        MyTable.create_table().run_sync()

    def tearDown(self):
        MyTable.alter().drop_table().run_sync()

    def test_bytea(self):
        """
        Test storing a valid bytes value.
        """
        row = MyTable(token=b"my-token")
        row.save().run_sync()
        self.assertEqual(row.token, b"my-token")

        self.assertEqual(
            MyTable.select(MyTable.token).first().run_sync(),
            {"token": b"my-token"},
        )


class TestByteaDefault(TestCase):
    def setUp(self):
        MyTableDefault.create_table().run_sync()

    def tearDown(self):
        MyTableDefault.alter().drop_table().run_sync()

    def test_json_default(self):
        row = MyTableDefault()
        row.save().run_sync()

        self.assertEqual(row.token, b"")
        self.assertEqual(row.token_bytes, b"my-token")
        self.assertEqual(row.token_bytearray, b"my-token")
        self.assertEqual(row.token_none, None)

    def test_invalid_default(self):
        with self.assertRaises(ValueError):
            for value in ("a", 1, ("x", "y", "z")):
                Bytea(default=value)  # type: ignore



from unittest import TestCase
from unittest.mock import MagicMock, patch

from piccolo.apps.user.commands.change_permissions import (
    Level,
    change_permissions,
)
from piccolo.apps.user.tables import BaseUser
from piccolo.utils.sync import run_sync


class TestChangePassword(TestCase):
    def setUp(self):
        BaseUser.create_table(if_not_exists=True).run_sync()

        BaseUser(
            username="bob",
            password="bob123",
            first_name="Bob",
            last_name="Jones",
            email="bob@gmail.com",
            active=False,
            admin=False,
            superuser=False,
        ).save().run_sync()

    def tearDown(self):
        BaseUser.alter().drop_table().run_sync()

    @patch("piccolo.apps.user.commands.change_permissions.colored_string")
    def test_user_doesnt_exist(self, colored_string: MagicMock):
        run_sync(change_permissions(username="sally"))
        colored_string.assert_called_once_with(
            "User sally doesn't exist!", level=Level.medium
        )

    def test_admin(self):
        run_sync(change_permissions(username="bob", admin=True))
        self.assertTrue(
            BaseUser.exists()
            .where(BaseUser.username == "bob", BaseUser.admin.eq(True))
            .run_sync()
        )

    def test_active(self):
        run_sync(change_permissions(username="bob", active=True))
        self.assertTrue(
            BaseUser.exists()
            .where(BaseUser.username == "bob", BaseUser.active.eq(True))
            .run_sync()
        )

    def test_superuser(self):
        run_sync(change_permissions(username="bob", superuser=True))
        self.assertTrue(
            BaseUser.exists()
            .where(BaseUser.username == "bob", BaseUser.superuser.eq(True))
            .run_sync()
        )

    @patch("piccolo.apps.user.commands.change_permissions.colored_string")
    def test_no_params(self, colored_string):
        run_sync(change_permissions(username="bob"))
        colored_string.assert_called_once_with(
            "No changes detected", level=Level.medium
        )



from piccolo.columns.column_types import Integer, Serial, Varchar
from piccolo.table import Table
from tests.base import DBTestCase, engine_is, engines_only, engines_skip


class Band(Table):
    id: Serial
    name = Varchar(db_column_name="regrettable_column_name")
    popularity = Integer()


class TestDBColumnName(DBTestCase):
    """
    By using the ``db_column_name`` arg, the user can map a ``Column`` to a
    database column with a different name. For example:

    .. code-block:: python

        class MyTable(Table):
            class_ = Varchar(db_column_name='class')

    """

    def setUp(self):
        Band.create_table().run_sync()

    def tearDown(self):
        Band.alter().drop_table().run_sync()

    @engines_only("postgres", "cockroach")
    def test_column_name_correct(self):
        """
        Make sure the column has the correct name in the database.
        """
        self.get_postgres_column_definition(
            tablename="band", column_name="regrettable_column_name"
        )

        with self.assertRaises(ValueError):
            self.get_postgres_column_definition(
                tablename="band", column_name="name"
            )

    def test_save(self):
        """
        Make sure save queries work correctly.
        """
        band = Band(name="Pythonistas", popularity=1000)
        band.save().run_sync()

        band_from_db = Band.objects().first().run_sync()
        assert band_from_db is not None
        self.assertEqual(band_from_db.name, "Pythonistas")

    def test_create(self):
        """
        Make sure create queries work correctly.
        """
        band = (
            Band.objects()
            .create(name="Pythonistas", popularity=1000)
            .run_sync()
        )
        self.assertEqual(band.name, "Pythonistas")

        band_from_db = Band.objects().first().run_sync()
        assert band_from_db is not None
        self.assertEqual(band_from_db.name, "Pythonistas")

    def test_select(self):
        """
        Make sure that select queries just return what is stored in the
        database. We might add an option in the future which maps the column
        name to it's alias, but it's hard to predict what behaviour the user
        wants.
        """
        Band.objects().create(name="Pythonistas", popularity=1000).run_sync()

        # Make sure we can select all columns
        bands = Band.select().run_sync()
        if engine_is("cockroach"):
            self.assertEqual(
                bands,
                [
                    {
                        "id": bands[0]["id"],
                        "regrettable_column_name": "Pythonistas",
                        "popularity": 1000,
                    }
                ],
            )
        else:
            self.assertEqual(
                bands,
                [
                    {
                        "id": 1,
                        "regrettable_column_name": "Pythonistas",
                        "popularity": 1000,
                    }
                ],
            )

        # Make sure we can select a single column
        bands = Band.select(Band.name).run_sync()
        self.assertEqual(
            bands,
            [
                {
                    "regrettable_column_name": "Pythonistas",
                }
            ],
        )

        # Make sure aliases still work
        bands = Band.select(Band.name.as_alias("name")).run_sync()
        self.assertEqual(
            bands,
            [
                {
                    "name": "Pythonistas",
                }
            ],
        )

    def test_update(self):
        """
        Make sure update queries work correctly.
        """
        Band.objects().create(name="Pythonistas", popularity=1000).run_sync()

        Band.update({Band.name: "Pythonistas 2"}, force=True).run_sync()

        bands = Band.select().run_sync()
        if engine_is("cockroach"):
            self.assertEqual(
                bands,
                [
                    {
                        "id": bands[0]["id"],
                        "regrettable_column_name": "Pythonistas 2",
                        "popularity": 1000,
                    }
                ],
            )
        else:
            self.assertEqual(
                bands,
                [
                    {
                        "id": 1,
                        "regrettable_column_name": "Pythonistas 2",
                        "popularity": 1000,
                    }
                ],
            )

        Band.update({"name": "Pythonistas 3"}, force=True).run_sync()

        bands = Band.select().run_sync()
        if engine_is("cockroach"):
            self.assertEqual(
                bands,
                [
                    {
                        "id": bands[0]["id"],
                        "regrettable_column_name": "Pythonistas 3",
                        "popularity": 1000,
                    }
                ],
            )
        else:
            self.assertEqual(
                bands,
                [
                    {
                        "id": 1,
                        "regrettable_column_name": "Pythonistas 3",
                        "popularity": 1000,
                    }
                ],
            )

    @engines_skip("cockroach")
    def test_delete(self):
        """
        Make sure delete queries work correctly.
        """
        Band.insert(
            Band(name="Pythonistas", popularity=1000),
            Band(name="Rustaceans", popularity=500),
        ).run_sync()

        bands = Band.select().run_sync()
        self.assertEqual(
            bands,
            [
                {
                    "id": 1,
                    "regrettable_column_name": "Pythonistas",
                    "popularity": 1000,
                },
                {
                    "id": 2,
                    "regrettable_column_name": "Rustaceans",
                    "popularity": 500,
                },
            ],
        )

        Band.delete().where(Band.name == "Rustaceans").run_sync()

        bands = Band.select().run_sync()
        self.assertEqual(
            bands,
            [
                {
                    "id": 1,
                    "regrettable_column_name": "Pythonistas",
                    "popularity": 1000,
                }
            ],
        )

    @engines_only("cockroach")
    def test_delete_alt(self):
        """
        Make sure delete queries work correctly.
        """
        result = (
            Band.insert(
                Band(name="Pythonistas", popularity=1000),
                Band(name="Rustaceans", popularity=500),
            )
            .returning(Band.id)
            .run_sync()
        )

        bands = Band.select().run_sync()
        self.assertEqual(
            bands,
            [
                {
                    "id": result[0]["id"],
                    "regrettable_column_name": "Pythonistas",
                    "popularity": 1000,
                },
                {
                    "id": result[1]["id"],
                    "regrettable_column_name": "Rustaceans",
                    "popularity": 500,
                },
            ],
        )

        Band.delete().where(Band.name == "Rustaceans").run_sync()

        bands = Band.select().run_sync()
        self.assertEqual(
            bands,
            [
                {
                    "id": result[0]["id"],
                    "regrettable_column_name": "Pythonistas",
                    "popularity": 1000,
                }
            ],
        )



from __future__ import annotations

import os
import sys
import typing as t

from piccolo.apps.migrations.auto.migration_manager import MigrationManager
from piccolo.apps.migrations.commands.base import (
    BaseMigrationManager,
    MigrationResult,
)
from piccolo.apps.migrations.tables import Migration
from piccolo.conf.apps import AppConfig, MigrationModule
from piccolo.utils.printing import print_heading


class BackwardsMigrationManager(BaseMigrationManager):
    def __init__(
        self,
        app_name: str,
        migration_id: str,
        auto_agree: bool = False,
        clean: bool = False,
        preview: bool = False,
    ):
        self.migration_id = migration_id
        self.app_name = app_name
        self.auto_agree = auto_agree
        self.clean = clean
        self.preview = preview
        super().__init__()

    async def run_migrations_backwards(self, app_config: AppConfig):
        migration_modules: t.Dict[str, MigrationModule] = (
            self.get_migration_modules(app_config.migrations_folder_path)
        )

        ran_migration_ids = await Migration.get_migrations_which_ran(
            app_name=self.app_name
        )
        if len(ran_migration_ids) == 0:
            # Make sure a success is returned, as we don't want this
            # to appear as an error in automated scripts.
            message = "🏁 No migrations to reverse!"
            print(message)
            return MigrationResult(success=True, message=message)

        #######################################################################

        if self.migration_id == "all":
            earliest_migration_id = ran_migration_ids[0]
        elif self.migration_id == "1":
            earliest_migration_id = ran_migration_ids[-1]
        else:
            earliest_migration_id = self.migration_id

        if earliest_migration_id not in ran_migration_ids:
            message = (
                "Unrecognized migration name - must be one of "
                f"{ran_migration_ids}"
            )
            print(message, file=sys.stderr)
            return MigrationResult(success=False, message=message)

        #######################################################################

        latest_migration_id = ran_migration_ids[-1]

        start_index = ran_migration_ids.index(earliest_migration_id)
        end_index = ran_migration_ids.index(latest_migration_id) + 1

        subset = ran_migration_ids[start_index:end_index]
        reversed_migration_ids = list(reversed(subset))

        #######################################################################

        n = len(reversed_migration_ids)
        _continue = (
            "y"
            if self.auto_agree
            else input(
                f"Reverse {n} migration{'s' if n != 1 else ''}? [y/N] "
            ).lower()
        )
        if _continue == "y":
            for migration_id in reversed_migration_ids:
                migration_module = migration_modules[migration_id]
                response = await migration_module.forwards()

                if isinstance(response, MigrationManager):
                    if self.preview:
                        response.preview = True
                    await response.run(backwards=True)
                if not self.preview:
                    await Migration.delete().where(
                        Migration.name == migration_id
                    ).run()

                    if self.clean and migration_module.__file__:
                        os.unlink(migration_module.__file__)

                print("ok! ✔️")
            return MigrationResult(success=True)

        else:  # pragma: no cover
            message = "Not proceeding."
            print(message, file=sys.stderr)
            return MigrationResult(success=False, message=message)

    async def run(self) -> MigrationResult:
        await self.create_migration_table()
        app_config = self.get_app_config(self.app_name)
        return await self.run_migrations_backwards(app_config=app_config)


async def run_backwards(
    app_name: str,
    migration_id: str = "1",
    auto_agree: bool = False,
    clean: bool = False,
    preview: bool = False,
) -> MigrationResult:
    if app_name == "all":
        sorted_app_names = BaseMigrationManager().get_sorted_app_names()
        sorted_app_names.reverse()

        names = [f"'{name}'" for name in sorted_app_names]
        _continue = (
            "y"
            if auto_agree
            else input(
                "You are about to undo the migrations for the following "
                "apps:\n"
                f"{', '.join(names)}\n"
                "Are you sure you want to continue? [y/N] "
            ).lower()
        )

        if _continue != "y":
            return MigrationResult(success=False, message="user cancelled")
        for _app_name in sorted_app_names:
            print_heading(_app_name)
            manager = BackwardsMigrationManager(
                app_name=_app_name,
                migration_id="all",
                auto_agree=auto_agree,
                preview=preview,
            )
            await manager.run()
        return MigrationResult(success=True)
    else:
        manager = BackwardsMigrationManager(
            app_name=app_name,
            migration_id=migration_id,
            auto_agree=auto_agree,
            clean=clean,
            preview=preview,
        )
        return await manager.run()


async def backwards(
    app_name: str,
    migration_id: str = "1",
    auto_agree: bool = False,
    clean: bool = False,
    preview: bool = False,
):
    """
    Undo migrations up to a specific migration.

    :param app_name:
        The app to reverse migrations for. Specify a value of 'all' to reverse
        migrations for all apps.
    :param migration_id:
        Migrations will be reversed up to and including this migration_id.
        Specify a value of 'all' to undo all of the migrations. Specify a
        value of '1' to undo the most recent migration.
    :param auto_agree:
        If true, automatically agree to any input prompts.
    :param clean:
        If true, the migration files which have been run backwards are deleted
        from the disk after completing.
    :param preview:
        If true, don't actually run the migration, just print the SQL queries.

    """
    response = await run_backwards(
        app_name=app_name,
        migration_id=migration_id,
        auto_agree=auto_agree,
        clean=clean,
        preview=preview,
    )

    if not response.success:
        sys.exit(1)



from enum import Enum

from piccolo.columns import (
    JSON,
    JSONB,
    BigInt,
    ForeignKey,
    Integer,
    Numeric,
    Serial,
    Text,
    Varchar,
)
from piccolo.columns.readable import Readable
from piccolo.engine.finder import engine_finder
from piccolo.table import Table

engine = engine_finder()

###############################################################################
# Simple example


class Manager(Table):
    id: Serial
    name = Varchar(length=50)

    @classmethod
    def get_readable(cls) -> Readable:
        return Readable(template="%s", columns=[cls.name])


class Band(Table):
    id: Serial
    name = Varchar(length=50)
    manager = ForeignKey(Manager, null=True)
    popularity = (
        BigInt(default=0)
        if engine and engine.engine_type == "cockroach"
        else Integer(default=0)
    )

    @classmethod
    def get_readable(cls) -> Readable:
        return Readable(template="%s", columns=[cls.name])


###############################################################################
# More complex


class Venue(Table):
    id: Serial
    name = Varchar(length=100)
    capacity = Integer(default=0, secret=True)

    @classmethod
    def get_readable(cls) -> Readable:
        return Readable(template="%s", columns=[cls.name])


class Concert(Table):
    id: Serial
    band_1 = ForeignKey(Band)
    band_2 = ForeignKey(Band)
    venue = ForeignKey(Venue)

    @classmethod
    def get_readable(cls) -> Readable:
        return Readable(
            template="%s and %s at %s, capacity %s",
            columns=[
                cls.band_1.name,
                cls.band_2.name,
                cls.venue.name,
                cls.venue.capacity,
            ],
        )


class Ticket(Table):
    id: Serial
    concert = ForeignKey(Concert)
    price = Numeric(digits=(5, 2))


class Poster(Table, tags=["special"]):
    """
    Has tags for tests which need it.
    """

    id: Serial
    content = Text()


class Shirt(Table):
    """
    Used for testing columns with a choices attribute.
    """

    class Size(str, Enum):
        small = "s"
        medium = "m"
        large = "l"

    id: Serial
    size = Varchar(length=1, choices=Size, default=Size.large)


class RecordingStudio(Table):
    """
    Used for testing JSON and JSONB columns.
    """

    id: Serial
    facilities = JSON()
    facilities_b = JSONB()


class Instrument(Table):
    """
    Used for testing foreign keys to a table with a JSON column.
    """

    id: Serial
    name = Varchar()
    recording_studio = ForeignKey(RecordingStudio)



import asyncio

from viztracer import VizTracer

from piccolo.columns.column_types import Varchar
from piccolo.engine.postgres import PostgresEngine
from piccolo.table import Table

DB = PostgresEngine(config={"database": "piccolo_profile"})


class Band(Table, db=DB):
    name = Varchar()


async def setup():
    await Band.alter().drop_table(if_exists=True)
    await Band.create_table(if_not_exists=True)
    await Band.insert(*[Band(name="test") for _ in range(1000)])


class Trace:
    def __enter__(self):
        self.tracer = VizTracer(log_async=True)
        self.tracer.start()

    def __exit__(self, *args):
        self.tracer.stop()
        self.tracer.save()


async def run_queries():
    await setup()

    with Trace():
        await Band.select()


if __name__ == "__main__":
    asyncio.run(run_queries())



import unittest
from enum import Enum

from piccolo.testing.random_builder import RandomBuilder


class TestRandomBuilder(unittest.TestCase):
    def test_next_bool(self):
        random_bool = RandomBuilder.next_bool()
        self.assertIn(random_bool, [True, False])

    def test_next_bytes(self):
        random_bytes = RandomBuilder.next_bytes(length=100)
        self.assertEqual(len(random_bytes), 100)

    def test_next_date(self):
        random_date = RandomBuilder.next_date()
        self.assertGreaterEqual(random_date.year, 2000)
        self.assertLessEqual(random_date.year, 2050)

    def test_next_datetime(self):
        random_datetime = RandomBuilder.next_datetime()
        self.assertGreaterEqual(random_datetime.year, 2000)
        self.assertLessEqual(random_datetime.year, 2050)

    def test_next_enum(self):
        class Color(Enum):
            RED = 1
            BLUE = 2

        random_enum = RandomBuilder.next_enum(Color)
        self.assertIsInstance(random_enum, int)

    def test_next_float(self):
        random_float = RandomBuilder.next_float(maximum=1000)
        self.assertLessEqual(random_float, 1000)

    def test_next_int(self):
        random_int = RandomBuilder.next_int()
        self.assertLessEqual(random_int, 2147483647)

    def test_next_str(self):
        random_str = RandomBuilder.next_str(length=64)
        self.assertLessEqual(len(random_str), 64)

    def test_next_time(self):
        RandomBuilder.next_time()

    def test_next_timedelta(self):
        random_timedelta = RandomBuilder.next_timedelta()
        self.assertLessEqual(random_timedelta.days, 7)

    def test_next_uuid(self):
        RandomBuilder.next_uuid()



import datetime
from functools import partial
from unittest import TestCase

from piccolo.columns.column_types import Time
from piccolo.columns.defaults.time import TimeNow
from piccolo.table import Table
from tests.base import engines_skip


class MyTable(Table):
    created_on = Time()


class MyTableDefault(Table):
    created_on = Time(default=TimeNow())


class TestTime(TestCase):
    def setUp(self):
        MyTable.create_table().run_sync()

    def tearDown(self):
        MyTable.alter().drop_table().run_sync()

    @engines_skip("cockroach")
    def test_timestamp(self):
        created_on = datetime.datetime.now().time()
        row = MyTable(created_on=created_on)
        row.save().run_sync()

        result = MyTable.objects().first().run_sync()
        assert result is not None
        self.assertEqual(result.created_on, created_on)


class TestTimeDefault(TestCase):
    def setUp(self):
        MyTableDefault.create_table().run_sync()

    def tearDown(self):
        MyTableDefault.alter().drop_table().run_sync()

    @engines_skip("cockroach")
    def test_timestamp(self):
        created_on = datetime.datetime.now().time()
        row = MyTableDefault()
        row.save().run_sync()

        _datetime = partial(datetime.datetime, year=2020, month=1, day=1)

        result = MyTableDefault.objects().first().run_sync()
        assert result is not None
        self.assertLess(
            _datetime(
                hour=result.created_on.hour,
                minute=result.created_on.minute,
                second=result.created_on.second,
            )
            - _datetime(
                hour=created_on.hour,
                minute=created_on.minute,
                second=created_on.second,
            ),
            datetime.timedelta(seconds=1),
        )



from __future__ import annotations

import sys
import typing as t

import typing_extensions

from piccolo.apps.fixtures.commands.shared import (
    FixtureConfig,
    create_pydantic_fixture_model,
)
from piccolo.conf.apps import Finder
from piccolo.engine import engine_finder
from piccolo.query.mixins import OnConflictAction
from piccolo.table import Table, sort_table_classes
from piccolo.utils.encoding import load_json
from piccolo.utils.list import batch


async def load_json_string(
    json_string: str,
    chunk_size: int = 1000,
    on_conflict_action: t.Optional[OnConflictAction] = None,
):
    """
    Parses the JSON string, and inserts the parsed data into the database.
    """
    # We have to deserialise the JSON to find out which apps and tables it
    # contains, so we can create a Pydantic model.
    # Then we let Pydantic do the proper deserialisation, as it does a much
    # better job of deserialising dates, datetimes, bytes etc.
    deserialised_contents = load_json(json_string)

    app_names = deserialised_contents.keys()

    fixture_configs = [
        FixtureConfig(
            app_name=app_name,
            table_class_names=list(deserialised_contents[app_name].keys()),
        )
        for app_name in app_names
    ]
    pydantic_model_class = create_pydantic_fixture_model(
        fixture_configs=fixture_configs
    )

    fixture_pydantic_model = pydantic_model_class.model_validate_json(
        json_string
    )

    finder = Finder()
    engine = engine_finder()

    if not engine:
        raise Exception("Unable to find the engine.")

    # This is what we want to the insert into the database:
    data: t.Dict[t.Type[Table], t.List[Table]] = {}

    for app_name in app_names:
        app_model = getattr(fixture_pydantic_model, app_name)

        for (
            table_class_name,
            model_instance_list,
        ) in app_model.__dict__.items():
            table_class = finder.get_table_with_name(
                app_name, table_class_name
            )
            data[table_class] = [
                table_class.from_dict(row.__dict__)
                for row in model_instance_list
            ]

    # We have to sort the table classes based on foreign key, so we insert
    # the data in the right order.
    sorted_table_classes = sort_table_classes(list(data.keys()))

    async with engine.transaction():
        for table_class in sorted_table_classes:
            rows = data[table_class]

            for chunk in batch(data=rows, chunk_size=chunk_size):
                query = table_class.insert(*chunk)
                if on_conflict_action is not None:
                    query = query.on_conflict(
                        target=table_class._meta.primary_key,
                        action=on_conflict_action,
                        values=table_class._meta.columns,
                    )
                await query.run()


async def load(
    path: str = "fixture.json",
    chunk_size: int = 1000,
    on_conflict: t.Optional[
        typing_extensions.Literal["DO NOTHING", "DO UPDATE"]
    ] = None,
):
    """
    Reads the fixture file, and loads the contents into the database.

    :param path:
        The path of the fixture file.

    :param chunk_size:
        The maximum number of rows to insert at a time. This is usually
        determined by the database adapter, which has a max number of
        parameters per query.

    :param on_conflict:
        If specified, the fixture will be upserted, meaning that if a row
        already exists with a matching primary key, then it will be overridden
        if "DO UPDATE", or it will be ignored if "DO NOTHING".

    """
    with open(path, "r") as f:
        contents = f.read()

    on_conflict_action: t.Optional[OnConflictAction] = None

    if on_conflict:
        try:
            on_conflict_action = OnConflictAction(on_conflict.upper())
        except ValueError:
            sys.exit(
                f"{on_conflict} isn't a valid option - use 'DO NOTHING' or "
                "'DO UPDATE'."
            )

    await load_json_string(
        contents,
        chunk_size=chunk_size,
        on_conflict_action=on_conflict_action,
    )



from unittest.mock import patch

from tests.base import DBTestCase
from tests.example_apps.music.tables import Manager


class TestLogging(DBTestCase):
    def tearDown(self):
        Manager._meta.db.log_queries = False
        Manager._meta.db.log_responses = False
        super().tearDown()

    def test_log_queries(self):
        Manager._meta.db.log_queries = True

        with patch("piccolo.engine.base.Engine.print_query") as print_query:
            Manager.select().run_sync()
            print_query.assert_called_once()

    def test_log_responses(self):
        Manager._meta.db.log_responses = True

        with patch(
            "piccolo.engine.base.Engine.print_response"
        ) as print_response:
            Manager.select().run_sync()
            print_response.assert_called_once()

    def test_log_queries_and_responses(self):
        Manager._meta.db.log_queries = True
        Manager._meta.db.log_responses = True

        with patch("piccolo.engine.base.Engine.print_query") as print_query:
            with patch(
                "piccolo.engine.base.Engine.print_response"
            ) as print_response:
                Manager.select().run_sync()
                print_query.assert_called_once()
                print_response.assert_called_once()



from unittest import TestCase

from piccolo.query.functions.string import Reverse, Upper
from piccolo.querystring import QueryString
from piccolo.table import create_db_tables_sync, drop_db_tables_sync
from tests.base import engines_skip
from tests.example_apps.music.tables import Band, Manager


class FunctionTest(TestCase):
    tables = (Band, Manager)

    def setUp(self) -> None:
        create_db_tables_sync(*self.tables)

        manager = Manager({Manager.name: "Guido"})
        manager.save().run_sync()

        band = Band({Band.name: "Pythonistas", Band.manager: manager})
        band.save().run_sync()

    def tearDown(self) -> None:
        drop_db_tables_sync(*self.tables)


class TestUpperFunction(FunctionTest):

    def test_column(self):
        """
        Make sure we can uppercase a column's value.
        """
        response = Band.select(Upper(Band.name)).run_sync()
        self.assertListEqual(response, [{"upper": "PYTHONISTAS"}])

    def test_alias(self):
        response = Band.select(Upper(Band.name, alias="name")).run_sync()
        self.assertListEqual(response, [{"name": "PYTHONISTAS"}])

    def test_joined_column(self):
        """
        Make sure we can uppercase a column's value from a joined table.
        """
        response = Band.select(Upper(Band.manager._.name)).run_sync()
        self.assertListEqual(response, [{"upper": "GUIDO"}])


@engines_skip("sqlite")
class TestNested(FunctionTest):
    """
    Skip the the test for SQLite, as it doesn't support ``Reverse``.
    """

    def test_nested(self):
        """
        Make sure we can nest functions.
        """
        response = Band.select(Upper(Reverse(Band.name))).run_sync()
        self.assertListEqual(response, [{"upper": "SATSINOHTYP"}])

    def test_nested_with_joined_column(self):
        """
        Make sure nested functions can be used on a column from a joined table.
        """
        response = Band.select(Upper(Reverse(Band.manager._.name))).run_sync()
        self.assertListEqual(response, [{"upper": "ODIUG"}])

    def test_nested_within_querystring(self):
        """
        If we wrap a function in a custom QueryString - make sure the columns
        are still accessible, so joins are successful.
        """
        response = Band.select(
            QueryString("CONCAT({}, '!')", Upper(Band.manager._.name)),
        ).run_sync()

        self.assertListEqual(response, [{"concat": "GUIDO!"}])


class TestWhereClause(FunctionTest):

    def test_where(self):
        """
        Make sure where clauses work with functions.
        """
        response = (
            Band.select(Band.name)
            .where(Upper(Band.name) == "PYTHONISTAS")
            .run_sync()
        )
        self.assertListEqual(response, [{"name": "Pythonistas"}])

    def test_where_with_joined_column(self):
        """
        Make sure where clauses work with functions, when a joined column is
        used.
        """
        response = (
            Band.select(Band.name)
            .where(Upper(Band.manager._.name) == "GUIDO")
            .run_sync()
        )
        self.assertListEqual(response, [{"name": "Pythonistas"}])



import secrets
from unittest import TestCase
from unittest.mock import MagicMock, call, patch

from piccolo.apps.user.tables import BaseUser


class TestCreateUserTable(TestCase):
    def test_create_user_table(self):
        """
        Make sure the table can be created.
        """
        exception = None
        try:
            BaseUser.create_table().run_sync()
        except Exception as e:
            exception = e
        else:
            BaseUser.alter().drop_table().run_sync()

        if exception:
            raise exception

        self.assertFalse(exception)


class TestInstantiateUser(TestCase):
    def setUp(self):
        BaseUser.create_table().run_sync()

    def tearDown(self):
        BaseUser.alter().drop_table().run_sync()

    def test_valid_credentials(self):
        BaseUser(username="bob", password="abc123%£1pscl")

    def test_malicious_password(self):
        malicious_password = secrets.token_urlsafe(1000)
        with self.assertRaises(ValueError) as manager:
            BaseUser(username="bob", password=malicious_password)
        self.assertEqual(
            manager.exception.__str__(), "The password is too long."
        )


class TestLogin(TestCase):
    def setUp(self):
        BaseUser.create_table().run_sync()

    def tearDown(self):
        BaseUser.alter().drop_table().run_sync()

    @patch("piccolo.apps.user.tables.logger")
    def test_login(self, logger: MagicMock):
        username = "bob"
        password = "Bob123$$$"
        email = "bob@bob.com"

        user = BaseUser(username=username, password=password, email=email)
        user.save().run_sync()

        # Test correct password
        authenticated = BaseUser.login_sync(username, password)
        self.assertTrue(authenticated == user.id)

        # Test incorrect password
        authenticated = BaseUser.login_sync(username, "blablabla")
        self.assertTrue(authenticated is None)

        # Test ultra long password
        malicious_password = secrets.token_urlsafe(1000)
        authenticated = BaseUser.login_sync(username, malicious_password)
        self.assertTrue(authenticated is None)
        self.assertEqual(
            logger.method_calls,
            [call.warning("Excessively long password provided.")],
        )

        # Test ulta long username
        logger.reset_mock()
        malicious_username = secrets.token_urlsafe(1000)
        authenticated = BaseUser.login_sync(malicious_username, password)
        self.assertTrue(authenticated is None)
        self.assertEqual(
            logger.method_calls,
            [call.warning("Excessively long username provided.")],
        )

    def test_update_password(self):
        username = "bob"
        password = "Bob123$$$"
        email = "bob@bob.com"

        user = BaseUser(username=username, password=password, email=email)
        user.save().run_sync()

        authenticated = BaseUser.login_sync(username, password)
        self.assertTrue(authenticated is not None)

        # Test success
        new_password = "XXX111"
        BaseUser.update_password_sync(username, new_password)
        authenticated = BaseUser.login_sync(username, new_password)
        self.assertTrue(authenticated is not None)

        # Test ultra long password
        malicious_password = secrets.token_urlsafe(1000)
        with self.assertRaises(ValueError) as manager:
            BaseUser.update_password_sync(username, malicious_password)
        self.assertEqual(
            manager.exception.__str__(),
            "The password is too long.",
        )

        # Test short passwords
        short_password = "abc"
        with self.assertRaises(ValueError) as manager:
            BaseUser.update_password_sync(username, short_password)
        self.assertEqual(
            manager.exception.__str__(),
            "The password is too short.",
        )

        # Test no password
        empty_password = ""
        with self.assertRaises(ValueError) as manager:
            BaseUser.update_password_sync(username, empty_password)
        self.assertEqual(
            manager.exception.__str__(),
            "A password must be provided.",
        )

        # Test hashed password
        hashed_password = "pbkdf2_sha256$abc123"
        with self.assertRaises(ValueError) as manager:
            BaseUser.update_password_sync(username, hashed_password)
        self.assertEqual(
            manager.exception.__str__(),
            "Do not pass a hashed password.",
        )


class TestCreateUserFromFixture(TestCase):
    def setUp(self):
        BaseUser.create_table().run_sync()

    def tearDown(self):
        BaseUser.alter().drop_table().run_sync()

    def test_create_user_from_fixture(self):
        the_data = {
            "id": 2,
            "username": "",
            "password": "pbkdf2_sha256$10000$19ed2c0d6cbe0868a70be6"
            "446b93ed5b$c862974665ccc25b334ed42fa7e96a41"
            "04d5ddff0c2e56e0e5b1d0efc67e9d03",
            "first_name": "",
            "last_name": "",
            "email": "",
            "active": False,
            "admin": False,
            "superuser": False,
            "last_login": None,
        }
        user = BaseUser.from_dict(the_data)
        self.assertIsInstance(user, BaseUser)
        self.assertEqual(user.password, the_data["password"])


class TestCreateUser(TestCase):
    def setUp(self):
        BaseUser.create_table().run_sync()

    def tearDown(self):
        BaseUser.alter().drop_table().run_sync()

    def test_success(self):
        user = BaseUser.create_user_sync(username="bob", password="abc123")
        self.assertTrue(isinstance(user, BaseUser))
        self.assertEqual(
            BaseUser.login_sync(username="bob", password="abc123"), user.id
        )

    @patch("piccolo.apps.user.tables.logger")
    def test_hashed_password_error(self, logger: MagicMock):
        with self.assertRaises(ValueError) as manager:
            BaseUser.create_user_sync(
                username="bob", password="pbkdf2_sha256$10000"
            )

        self.assertEqual(
            manager.exception.__str__(), "Do not pass a hashed password."
        )
        self.assertEqual(
            logger.method_calls,
            [
                call.warning(
                    "Tried to create a user with an already hashed password."
                )
            ],
        )

    def test_short_password_error(self):
        with self.assertRaises(ValueError) as manager:
            BaseUser.create_user_sync(username="bob", password="abc")

        self.assertEqual(
            manager.exception.__str__(), "The password is too short."
        )

    def test_long_password_error(self):
        with self.assertRaises(ValueError) as manager:
            BaseUser.create_user_sync(
                username="bob",
                password="x" * (BaseUser._max_password_length + 1),
            )

        self.assertEqual(
            manager.exception.__str__(), "The password is too long."
        )

    def test_no_username_error(self):
        with self.assertRaises(ValueError) as manager:
            BaseUser.create_user_sync(
                username=None,  # type: ignore
                password="abc123",
            )

        self.assertEqual(
            manager.exception.__str__(), "A username must be provided."
        )

    def test_no_password_error(self):
        with self.assertRaises(ValueError) as manager:
            BaseUser.create_user_sync(
                username="bob",
                password=None,  # type: ignore
            )

        self.assertEqual(
            manager.exception.__str__(), "A password must be provided."
        )


class TestAutoHashingUpdate(TestCase):
    """
    Make sure that users with passwords which were hashed in earlier Piccolo
    versions are automatically re-hashed, meeting current best practices with
    the number of hashing iterations.
    """

    def setUp(self):
        BaseUser.create_table().run_sync()

    def tearDown(self):
        BaseUser.alter().drop_table().run_sync()

    def test_hash_update(self):
        # Create a user
        username = "bob"
        password = "abc123"
        user = BaseUser.create_user_sync(username=username, password=password)

        # Update their password, so it uses less than the recommended number
        # of hashing iterations.
        BaseUser.update(
            {
                BaseUser.password: BaseUser.hash_password(
                    password=password,
                    iterations=int(BaseUser._pbkdf2_iteration_count / 2),
                )
            }
        ).where(BaseUser.id == user.id).run_sync()

        # Login the user - Piccolo should detect their password needs rehashing
        # and update it.
        self.assertIsNotNone(
            BaseUser.login_sync(username=username, password=password)
        )

        user_data = (
            BaseUser.select(BaseUser.password)
            .where(BaseUser.id == user.id)
            .first()
            .run_sync()
        )
        assert user_data is not None
        hashed_password = user_data["password"]

        algorithm, iterations_, salt, hashed = BaseUser.split_stored_password(
            hashed_password
        )

        self.assertEqual(int(iterations_), BaseUser._pbkdf2_iteration_count)

        # Make sure subsequent logins work as expected
        self.assertIsNotNone(
            BaseUser.login_sync(username=username, password=password)
        )



from __future__ import annotations

import abc
import builtins
import datetime
import decimal
import inspect
import typing as t
import uuid
import warnings
from copy import deepcopy
from dataclasses import dataclass, field
from enum import Enum

from piccolo.columns import Column
from piccolo.columns.defaults.base import Default
from piccolo.columns.reference import LazyTableReference
from piccolo.table import Table
from piccolo.utils.repr import repr_class_instance

from .serialisation_legacy import deserialise_legacy_params

###############################################################################


class CanConflictWithGlobalNames(abc.ABC):
    @abc.abstractmethod
    def warn_if_is_conflicting_with_global_name(self): ...


class UniqueGlobalNamesMeta(type):
    """
    Metaclass for ``UniqueGlobalNames``.

    Fulfills the following functions:

    - Assure that no two class attributes have the same value.
    - Add class attributes `COLUMN_<capitalized column class name>`
      to the class for each column type.
    """

    def __new__(mcs, name, bases, class_attributes):
        class_attributes_with_columns = mcs.merge_class_attributes(
            class_attributes,
            mcs.get_column_class_attributes(),
        )

        return super().__new__(
            mcs,
            name,
            bases,
            mcs.merge_class_attributes(
                class_attributes_with_columns,
                {
                    "unique_names": mcs.get_unique_class_attribute_values(
                        class_attributes_with_columns
                    )
                },
            ),
        )

    @staticmethod
    def get_unique_class_attribute_values(
        class_attributes: t.Dict[str, t.Any]
    ) -> t.Set[t.Any]:
        """
        Return class attribute values.

        Raises an error if attribute values are not unique.
        """

        unique_attribute_values = set()
        for attribute_name, attribute_value in class_attributes.items():
            # Skip special attributes, i.e. "__<special attribute name>__"
            if attribute_name.startswith("__") and attribute_name.endswith(
                "__"
            ):
                continue

            if attribute_value in unique_attribute_values:
                raise ValueError(
                    f"Duplicate unique global name {attribute_value}"
                )
            unique_attribute_values.add(attribute_value)

        return unique_attribute_values

    @staticmethod
    def merge_class_attributes(
        class_attributes1: t.Dict[str, t.Any],
        class_attributes2: t.Dict[str, t.Any],
    ) -> t.Dict[str, t.Any]:
        """
        Merges two class attribute dictionaries.

        Raise an error if both dictionaries have an attribute
        with the same name.
        """

        for attribute_name in class_attributes2:
            if attribute_name in class_attributes1:
                raise ValueError(f"Duplicate class attribute {attribute_name}")

        return dict(**class_attributes1, **class_attributes2)

    @staticmethod
    def get_column_class_attributes() -> t.Dict[str, str]:
        """Automatically generates global names for each column type."""

        import piccolo.columns.column_types

        class_attributes: t.Dict[str, str] = {}
        for module_global in piccolo.columns.column_types.__dict__.values():
            try:
                if module_global is not Column and issubclass(
                    module_global, Column
                ):
                    class_attributes[
                        f"COLUMN_{module_global.__name__.upper()}"
                    ] = module_global.__name__
            except TypeError:
                pass

        return class_attributes


class UniqueGlobalNames(metaclass=UniqueGlobalNamesMeta):
    """
    Contains global names that may be used during serialisation.

    The global names are stored as class attributes. Names that may
    occur in the global namespace after serialisation should be listed here.

    This class is meant to prevent against the use of conflicting global
    names. If possible imports and global definitions should use this
    class to ensure that no conflicts arise during serialisation.
    """

    # Piccolo imports
    TABLE = Table.__name__
    DEFAULT = Default.__name__
    # Column types are omitted because they are added by metaclass

    # Standard library imports
    STD_LIB_ENUM = Enum.__name__
    STD_LIB_MODULE_DECIMAL = "decimal"

    # Third-party library imports
    EXTERNAL_MODULE_UUID = "uuid"
    EXTERNAL_UUID = f"{EXTERNAL_MODULE_UUID}.{uuid.UUID.__name__}"

    # This attribute is set in metaclass
    unique_names: t.Set[str]

    @classmethod
    def warn_if_is_conflicting_name(
        cls, name: str, name_type: str = "Name"
    ) -> None:
        """Raise an error if ``name`` matches a class attribute value."""

        if cls.is_conflicting_name(name):
            warnings.warn(
                f"{name_type} '{name}' could conflict with global name",
                UniqueGlobalNameConflictWarning,
            )

    @classmethod
    def is_conflicting_name(cls, name: str) -> bool:
        """Check if ``name`` matches a class attribute value."""

        return name in cls.unique_names

    @staticmethod
    def warn_if_are_conflicting_objects(
        objects: t.Iterable[CanConflictWithGlobalNames],
    ) -> None:
        """
        Call each object's ``raise_if_is_conflicting_with_global_name`` method.
        """

        for obj in objects:
            obj.warn_if_is_conflicting_with_global_name()


class UniqueGlobalNameConflictWarning(UserWarning):
    pass


###############################################################################


@dataclass
class Import(CanConflictWithGlobalNames):
    module: str
    target: t.Optional[str] = None
    expect_conflict_with_global_name: t.Optional[str] = None

    def __post_init__(self) -> None:
        if (
            self.expect_conflict_with_global_name is not None
            and not UniqueGlobalNames.is_conflicting_name(
                self.expect_conflict_with_global_name
            )
        ):
            raise ValueError(
                f"`expect_conflict_with_global_name="
                f'"{self.expect_conflict_with_global_name}"` '
                f"is not listed in `{UniqueGlobalNames.__name__}`"
            )

    def __repr__(self):
        if self.target is None:
            return f"import {self.module}"

        return f"from {self.module} import {self.target}"

    def __hash__(self):
        if self.target is None:
            return hash(f"{self.module}")

        return hash(f"{self.module}-{self.target}")

    def __lt__(self, other):
        return repr(self) < repr(other)

    def warn_if_is_conflicting_with_global_name(self):
        name = self.module if self.target is None else self.target
        if name == self.expect_conflict_with_global_name:
            return

        if UniqueGlobalNames.is_conflicting_name(name):
            UniqueGlobalNames.warn_if_is_conflicting_name(
                name, name_type="Import"
            )


class Definition(CanConflictWithGlobalNames, abc.ABC):
    @abc.abstractmethod
    def __repr__(self): ...

    ###########################################################################
    # To allow sorting:

    def __lt__(self, value):
        return self.__repr__() < value.__repr__()

    def __le__(self, value):
        return self.__repr__() <= value.__repr__()

    def __gt__(self, value):
        return self.__repr__() > value.__repr__()

    def __ge__(self, value):
        return self.__repr__() >= value.__repr__()


@dataclass
class SerialisedParams:
    params: t.Dict[str, t.Any]
    extra_imports: t.List[Import]
    extra_definitions: t.List[Definition] = field(default_factory=list)


###############################################################################


def check_equality(self, other):
    if getattr(other, "__hash__", None) is not None:
        return self.__hash__() == other.__hash__()
    else:
        return False


@dataclass
class SerialisedBuiltin:
    builtin: t.Any

    def __hash__(self):
        return hash(self.builtin.__name__)

    def __eq__(self, other):
        return check_equality(self, other)

    def __repr__(self):
        return self.builtin.__name__


@dataclass
class SerialisedClassInstance:
    instance: object

    def __hash__(self):
        return self.instance.__hash__()

    def __eq__(self, other):
        return check_equality(self, other)

    def __repr__(self):
        return repr_class_instance(self.instance)


@dataclass
class SerialisedColumnInstance:
    instance: Column
    serialised_params: SerialisedParams

    def __hash__(self):
        return self.instance.__hash__()

    def __eq__(self, other):
        return check_equality(self, other)

    def __repr__(self):
        args = ", ".join(
            f"{key}={self.serialised_params.params.get(key).__repr__()}"
            for key in self.instance._meta.params.keys()
        )

        return f"{self.instance.__class__.__name__}({args})"


@dataclass
class SerialisedEnumInstance:
    instance: Enum

    def __hash__(self):
        return hash(self.__repr__())

    def __eq__(self, other):
        return check_equality(self, other)

    def __repr__(self):
        return f"{self.instance.__class__.__name__}.{self.instance.name}"


@dataclass
class SerialisedTableType(Definition):
    table_type: t.Type[Table]

    def __hash__(self):
        return hash(
            f"{self.table_type._meta.tablename}-{self.table_type.__name__}"
        )

    def __eq__(self, other):
        return check_equality(self, other)

    def __repr__(self) -> str:
        tablename = self.table_type._meta.tablename

        # We have to add the primary key column definition too, so foreign
        # keys can be created with the correct type.
        pk_column = self.table_type._meta.primary_key
        pk_column_name = pk_column._meta.name
        serialised_pk_column = SerialisedColumnInstance(
            pk_column,
            serialised_params=serialise_params(params=pk_column._meta.params),
        )

        #######################################################################

        # When creating a ForeignKey, the user can specify a column other than
        # the primary key to reference.
        serialised_target_columns: t.Set[SerialisedColumnInstance] = set()

        for fk_column in self.table_type._meta._foreign_key_references:
            target_column = fk_column._foreign_key_meta.target_column
            if target_column is None:
                # Just references the primary key
                continue
            elif type(target_column) is str:
                column = self.table_type._meta.get_column_by_name(
                    target_column
                )
            elif isinstance(target_column, Column):
                column = self.table_type._meta.get_column_by_name(
                    target_column._meta.name
                )
            else:
                raise ValueError("Unrecognised `target_column` value.")

            serialised_target_columns.add(
                SerialisedColumnInstance(
                    column,
                    serialised_params=serialise_params(
                        params=column._meta.params
                    ),
                )
            )

        #######################################################################

        schema_str = (
            "None"
            if self.table_type._meta.schema is None
            else f'"{self.table_type._meta.schema}"'
        )

        definition = (
            f"class {self.table_class_name}"
            f'({UniqueGlobalNames.TABLE}, tablename="{tablename}", schema={schema_str}): '  # noqa: E501
            f"{pk_column_name} = {serialised_pk_column}"
        )

        for serialised_target_column in serialised_target_columns:
            definition += f"; {serialised_target_column.instance._meta.name} = {serialised_target_column}"  # noqa: E501

        return definition

    def __lt__(self, other):
        return repr(self) < repr(other)

    @property
    def table_class_name(self) -> str:
        return self.table_type.__name__

    def warn_if_is_conflicting_with_global_name(self) -> None:
        UniqueGlobalNames.warn_if_is_conflicting_name(self.table_class_name)


@dataclass
class SerialisedEnumType:
    enum_type: t.Type[Enum]

    def __hash__(self):
        return hash(self.__repr__())

    def __eq__(self, other):
        return check_equality(self, other)

    def __repr__(self):
        class_name = self.enum_type.__name__
        params = {i.name: i.value for i in self.enum_type}
        return f"{UniqueGlobalNames.STD_LIB_ENUM}('{class_name}', {params})"


@dataclass
class SerialisedCallable:
    callable_: t.Callable

    def __hash__(self):
        return hash(self.callable_.__name__)

    def __eq__(self, other):
        return check_equality(self, other)

    def __repr__(self):
        return self.callable_.__name__


@dataclass
class SerialisedUUID:
    instance: uuid.UUID

    def __hash__(self):
        return self.instance.int

    def __eq__(self, other):
        return check_equality(self, other)

    def __repr__(self):
        return f'{UniqueGlobalNames.EXTERNAL_UUID}("{str(self.instance)}")'


@dataclass
class SerialisedDecimal:
    instance: decimal.Decimal

    def __hash__(self):
        return hash(repr(self))

    def __eq__(self, other):
        return check_equality(self, other)

    def __repr__(self):
        return f"{UniqueGlobalNames.STD_LIB_MODULE_DECIMAL}." + repr(
            self.instance
        ).replace("'", '"')


###############################################################################


def serialise_params(params: t.Dict[str, t.Any]) -> SerialisedParams:
    """
    When writing column params to a migration file, we need to serialise some
    of the values.
    """
    params = deepcopy(params)
    extra_imports: t.List[Import] = []
    extra_definitions: t.List[Definition] = []

    for key, value in params.items():
        # Builtins, such as str, list and dict.
        if inspect.getmodule(value) == builtins:
            params[key] = SerialisedBuiltin(builtin=value)
            continue

        # Column instances
        if isinstance(value, Column):
            # For target_column (which is used by ForeignKey), we can just
            # serialise it as the column name:
            if key == "target_column":
                params[key] = value._meta.name
                continue

            ###################################################################

            # For Array definitions, we want to serialise the full column
            # definition:

            column: Column = value
            serialised_params: SerialisedParams = serialise_params(
                params=column._meta.params
            )

            # Include the extra imports and definitions required for the
            # column params.
            extra_imports.extend(serialised_params.extra_imports)
            extra_definitions.extend(serialised_params.extra_definitions)

            column_class_name = column.__class__.__name__
            extra_imports.append(
                Import(
                    module=column.__class__.__module__,
                    target=column_class_name,
                    expect_conflict_with_global_name=getattr(
                        UniqueGlobalNames,
                        f"COLUMN_{column_class_name.upper()}",
                        None,
                    ),
                )
            )
            params[key] = SerialisedColumnInstance(
                instance=value, serialised_params=serialised_params
            )
            continue

        # Class instances
        if isinstance(value, Default):
            params[key] = SerialisedClassInstance(instance=value)
            extra_imports.append(
                Import(
                    module=value.__class__.__module__,
                    target=value.__class__.__name__,
                    expect_conflict_with_global_name=UniqueGlobalNames.DEFAULT,
                )
            )
            continue

        # Dates and times
        if isinstance(
            value, (datetime.time, datetime.datetime, datetime.date)
        ):
            # Already has a good __repr__.
            extra_imports.append(
                Import(
                    module=value.__class__.__module__,
                    target=value.__class__.__name__,
                )
            )
            continue

        # UUIDs
        if isinstance(value, uuid.UUID):
            params[key] = SerialisedUUID(instance=value)
            extra_imports.append(
                Import(
                    module=UniqueGlobalNames.EXTERNAL_MODULE_UUID,
                    expect_conflict_with_global_name=(
                        UniqueGlobalNames.EXTERNAL_MODULE_UUID
                    ),
                )
            )
            continue

        # Decimals
        if isinstance(value, decimal.Decimal):
            params[key] = SerialisedDecimal(instance=value)
            extra_imports.append(
                Import(
                    module=UniqueGlobalNames.STD_LIB_MODULE_DECIMAL,
                    expect_conflict_with_global_name=(
                        UniqueGlobalNames.STD_LIB_MODULE_DECIMAL
                    ),
                )
            )
            continue

        # Enum instances
        if isinstance(value, Enum):
            if value.__module__.startswith("piccolo"):
                # It's an Enum defined within Piccolo, so we can safely import
                # it.
                params[key] = SerialisedEnumInstance(instance=value)
                extra_imports.append(
                    Import(
                        module=value.__module__,
                        target=value.__class__.__name__,
                    )
                )
            else:
                # It's a user defined Enum, so we'll insert the raw value.
                enum_serialised_params: SerialisedParams = serialise_params(
                    params={key: value.value}
                )
                params[key] = enum_serialised_params.params[key]
                extra_imports.extend(enum_serialised_params.extra_imports)
                extra_definitions.extend(
                    enum_serialised_params.extra_definitions
                )

            continue

        # Enum types
        if inspect.isclass(value) and issubclass(value, Enum):
            params[key] = SerialisedEnumType(enum_type=value)
            extra_imports.append(
                Import(
                    module="enum",
                    target=UniqueGlobalNames.STD_LIB_ENUM,
                    expect_conflict_with_global_name=(
                        UniqueGlobalNames.STD_LIB_ENUM
                    ),
                )
            )
            for member in value:
                type_ = type(member.value)
                module = inspect.getmodule(type_)

                if module and module != builtins:
                    module_name = module.__name__
                    extra_imports.append(
                        Import(module=module_name, target=type_.__name__)
                    )

        # Functions
        if inspect.isfunction(value):
            if value.__name__ == "<lambda>":
                raise ValueError("Lambdas can't be serialised")

            params[key] = SerialisedCallable(callable_=value)
            extra_imports.append(
                Import(module=value.__module__, target=value.__name__)
            )
            continue

        # Lazy imports - we need to resolve these now, in case the target
        # table class gets deleted in the future.
        if isinstance(value, LazyTableReference):
            table_type = value.resolve()
            params[key] = SerialisedCallable(callable_=table_type)
            extra_definitions.append(
                SerialisedTableType(table_type=table_type)
            )
            extra_imports.append(
                Import(
                    module=Table.__module__,
                    target=UniqueGlobalNames.TABLE,
                    expect_conflict_with_global_name=UniqueGlobalNames.TABLE,
                )
            )
            continue

        # Replace any Table class values into class and table names
        if inspect.isclass(value) and issubclass(value, Table):
            params[key] = SerialisedCallable(callable_=value)
            extra_definitions.append(SerialisedTableType(table_type=value))
            extra_imports.append(
                Import(
                    module=Table.__module__,
                    target=UniqueGlobalNames.TABLE,
                    expect_conflict_with_global_name=UniqueGlobalNames.TABLE,
                )
            )

            primary_key_class = value._meta.primary_key.__class__
            extra_imports.append(
                Import(
                    module=primary_key_class.__module__,
                    target=primary_key_class.__name__,
                    expect_conflict_with_global_name=getattr(
                        UniqueGlobalNames,
                        f"COLUMN_{primary_key_class.__name__.upper()}",
                        None,
                    ),
                )
            )
            # Include the extra imports and definitions required for the
            # primary column params.
            pk_serialised_params: SerialisedParams = serialise_params(
                params=value._meta.primary_key._meta.params
            )
            extra_imports.extend(pk_serialised_params.extra_imports)
            extra_definitions.extend(pk_serialised_params.extra_definitions)

            continue

        # Plain class type
        if inspect.isclass(value) and not issubclass(value, Enum):
            params[key] = SerialisedCallable(callable_=value)
            extra_imports.append(
                Import(module=value.__module__, target=value.__name__)
            )
            continue

        # All other types can remain as is.

    unique_extra_imports = list(set(extra_imports))
    UniqueGlobalNames.warn_if_are_conflicting_objects(unique_extra_imports)

    unique_extra_definitions = list(set(extra_definitions))
    UniqueGlobalNames.warn_if_are_conflicting_objects(unique_extra_definitions)

    return SerialisedParams(
        params=params,
        extra_imports=unique_extra_imports,
        extra_definitions=unique_extra_definitions,
    )


def deserialise_params(params: t.Dict[str, t.Any]) -> t.Dict[str, t.Any]:
    """
    When reading column params from a migration file, we need to convert
    them from their serialised form.
    """
    params = deepcopy(params)

    for key, value in params.items():
        # This is purely for backwards compatibility.
        if isinstance(value, str) and not isinstance(value, Enum):
            if value != "self":
                params[key] = deserialise_legacy_params(name=key, value=value)
        elif isinstance(value, SerialisedClassInstance):
            params[key] = value.instance
        elif isinstance(value, SerialisedUUID):
            params[key] = value.instance
        elif isinstance(value, SerialisedDecimal):
            params[key] = value.instance
        elif isinstance(value, SerialisedCallable):
            params[key] = value.callable_
        elif isinstance(value, SerialisedTableType):
            params[key] = value.table_type
        elif isinstance(value, SerialisedEnumType):
            params[key] = value.enum_type
        elif isinstance(value, SerialisedEnumInstance):
            params[key] = value.instance

    return params



import datetime
from unittest import TestCase

from piccolo.columns.column_types import Interval
from piccolo.columns.defaults.interval import IntervalCustom
from piccolo.table import Table


class MyTable(Table):
    interval = Interval()


class MyTableDefault(Table):
    interval = Interval(default=IntervalCustom(days=1))


class TestInterval(TestCase):
    def setUp(self):
        MyTable.create_table().run_sync()

    def tearDown(self):
        MyTable.alter().drop_table().run_sync()

    def test_interval(self):
        # Test a range of different timedeltas
        intervals = [
            datetime.timedelta(weeks=1),
            datetime.timedelta(days=1),
            datetime.timedelta(hours=1),
            datetime.timedelta(minutes=1),
            datetime.timedelta(seconds=1),
            datetime.timedelta(milliseconds=1),
            datetime.timedelta(microseconds=1),
        ]

        for interval in intervals:
            # Make sure that saving it works OK.
            row = MyTable(interval=interval)
            row.save().run_sync()

            # Make sure that fetching it back works OK.
            result = (
                MyTable.objects()
                .where(
                    MyTable._meta.primary_key
                    == getattr(row, MyTable._meta.primary_key._meta.name)
                )
                .first()
                .run_sync()
            )
            assert result is not None
            self.assertEqual(result.interval, interval)

    def test_interval_where_clause(self):
        """
        Make sure a range of where clauses resolve correctly.
        """
        interval = datetime.timedelta(hours=2)
        row = MyTable(interval=interval)
        row.save().run_sync()

        result = (
            MyTable.objects()
            .where(MyTable.interval < datetime.timedelta(hours=3))
            .first()
            .run_sync()
        )
        self.assertIsNotNone(result)

        result = (
            MyTable.objects()
            .where(MyTable.interval < datetime.timedelta(days=1))
            .first()
            .run_sync()
        )
        self.assertIsNotNone(result)

        result = (
            MyTable.objects()
            .where(MyTable.interval > datetime.timedelta(minutes=1))
            .first()
            .run_sync()
        )
        self.assertIsNotNone(result)

        result = (
            MyTable.exists()
            .where(MyTable.interval == datetime.timedelta(hours=2))
            .run_sync()
        )
        self.assertTrue(result)


class TestIntervalDefault(TestCase):
    def setUp(self):
        MyTableDefault.create_table().run_sync()

    def tearDown(self):
        MyTableDefault.alter().drop_table().run_sync()

    def test_interval(self):
        row = MyTableDefault()
        row.save().run_sync()

        result = MyTableDefault.objects().first().run_sync()
        assert result is not None
        self.assertEqual(result.interval.days, 1)



import json
from unittest import TestCase

from piccolo.table import create_db_tables_sync, drop_db_tables_sync
from tests.base import DBTestCase
from tests.example_apps.music.tables import Band, Instrument, RecordingStudio


class TestOutputList(DBTestCase):
    def test_output_as_list(self):
        self.insert_row()

        response = Band.select(Band.name).output(as_list=True).run_sync()
        self.assertEqual(response, ["Pythonistas"])

        # Make sure that if no rows are found, an empty list is returned.
        empty_response = (
            Band.select(Band.name)
            .where(Band.name == "ABC123")
            .output(as_list=True)
            .run_sync()
        )
        self.assertEqual(empty_response, [])


class TestOutputJSON(DBTestCase):
    def test_output_as_json(self):
        self.insert_row()

        response = Band.select(Band.name).output(as_json=True).run_sync()

        self.assertEqual(json.loads(response), [{"name": "Pythonistas"}])


class TestOutputLoadJSON(TestCase):
    tables = [RecordingStudio, Instrument]
    json = {"a": 123}

    def setUp(self):
        create_db_tables_sync(*self.tables)

        recording_studio = RecordingStudio(
            {
                RecordingStudio.facilities: self.json,
                RecordingStudio.facilities_b: self.json,
            }
        )
        recording_studio.save().run_sync()

        instrument = Instrument(
            {
                Instrument.recording_studio: recording_studio,
                Instrument.name: "Piccolo",
            }
        )
        instrument.save().run_sync()

    def tearDown(self):
        drop_db_tables_sync(*self.tables)

    def test_select(self):
        results = (
            RecordingStudio.select(
                RecordingStudio.facilities, RecordingStudio.facilities_b
            )
            .output(load_json=True)
            .run_sync()
        )

        self.assertEqual(
            results,
            [
                {
                    "facilities": self.json,
                    "facilities_b": self.json,
                }
            ],
        )

    def test_join(self):
        """
        Make sure it works correctly when the JSON column is on a joined table.

        https://github.com/piccolo-orm/piccolo/issues/1001

        """
        results = (
            Instrument.select(
                Instrument.name,
                Instrument.recording_studio._.facilities,
            )
            .output(load_json=True)
            .run_sync()
        )

        self.assertEqual(
            results,
            [
                {
                    "name": "Piccolo",
                    "recording_studio.facilities": self.json,
                }
            ],
        )

    def test_join_with_alias(self):
        results = (
            Instrument.select(
                Instrument.name,
                Instrument.recording_studio._.facilities.as_alias(
                    "facilities"
                ),
            )
            .output(load_json=True)
            .run_sync()
        )

        self.assertEqual(
            results,
            [
                {
                    "name": "Piccolo",
                    "facilities": self.json,
                }
            ],
        )

    def test_objects(self):
        results = RecordingStudio.objects().output(load_json=True).run_sync()
        self.assertEqual(results[0].facilities, self.json)
        self.assertEqual(results[0].facilities_b, self.json)


class TestOutputNested(DBTestCase):
    def test_output_nested(self):
        self.insert_row()

        response = (
            Band.select(Band.name, Band.manager.name)
            .output(nested=True)
            .run_sync()
        )
        self.assertEqual(
            response, [{"name": "Pythonistas", "manager": {"name": "Guido"}}]
        )

    def test_output_nested_with_first(self):
        self.insert_row()

        response = (
            Band.select(Band.name, Band.manager.name)
            .first()
            .output(nested=True)
            .run_sync()
        )
        assert response is not None
        self.assertDictEqual(
            response,  # type: ignore
            {"name": "Pythonistas", "manager": {"name": "Guido"}},
        )



from __future__ import annotations

import typing as t
from unittest import TestCase

import pytest

from piccolo.columns import BigInt, Integer, Numeric, Varchar
from piccolo.columns.base import Column
from piccolo.columns.column_types import ForeignKey, Text
from piccolo.schema import SchemaManager
from piccolo.table import Table
from tests.base import (
    DBTestCase,
    engine_version_lt,
    engines_only,
    is_running_sqlite,
)
from tests.example_apps.music.tables import Band, Manager


@pytest.mark.skipif(
    is_running_sqlite() and engine_version_lt(3.25),
    reason="SQLite version not supported",
)
class TestRenameColumn(DBTestCase):
    def _test_rename(
        self,
        existing_column: t.Union[Column, str],
        new_column_name: str = "rating",
    ):
        self.insert_row()

        rename_query = Band.alter().rename_column(
            existing_column, new_column_name
        )
        rename_query.run_sync()

        select_query = Band.raw("SELECT * FROM band")
        response = select_query.run_sync()

        column_names = response[0].keys()
        existing_column_name = (
            existing_column._meta.name
            if isinstance(existing_column, Column)
            else existing_column
        )
        self.assertTrue(
            (new_column_name in column_names)
            and (existing_column_name not in column_names)
        )

    def test_column(self):
        """
        Make sure a ``Column`` argument works.
        """
        self._test_rename(Band.popularity)

    def test_string(self):
        """
        Make sure a string argument works.
        """
        self._test_rename("popularity")

    def test_problematic_name(self):
        """
        Make sure we can rename columns with names which clash with SQL
        keywords.
        """
        self._test_rename(
            existing_column=Band.popularity, new_column_name="order"
        )


class TestRenameTable(DBTestCase):
    def test_rename(self):
        Band.alter().rename_table("act").run_sync()
        self.assertEqual(Band.raw("SELECT * FROM act").run_sync(), [])

    def tearDown(self):
        super().tearDown()
        self.run_sync("DROP TABLE IF EXISTS act")


@engines_only("postgres", "cockroach")
class TestDropColumn(DBTestCase):
    """
    Unfortunately this only works with Postgres at the moment.

    SQLite has very limited support for ALTER statements.
    """

    def _test_drop(self, column: t.Union[str, Column]):
        self.insert_row()

        Band.alter().drop_column(column).run_sync()

        response = Band.raw("SELECT * FROM band").run_sync()

        column_names = response[0].keys()
        self.assertNotIn("popularity", column_names)

    def test_drop_string(self):
        self._test_drop(Band.popularity)

    def test_drop_column(self):
        self._test_drop("popularity")


class TestAddColumn(DBTestCase):
    def _test_add_column(
        self, column: Column, column_name: str, expected_value: t.Any
    ):
        self.insert_row()
        Band.alter().add_column(column_name, column).run_sync()

        response = Band.raw("SELECT * FROM band").run_sync()

        column_names = response[0].keys()
        self.assertIn(column_name, column_names)

        self.assertEqual(response[0][column_name], expected_value)

    def test_integer(self):
        self._test_add_column(
            column=Integer(null=True, default=None),
            column_name="members",
            expected_value=None,
        )

    def test_foreign_key(self):
        self._test_add_column(
            column=ForeignKey(references=Manager),
            column_name="assistant_manager",
            expected_value=None,
        )

    def test_text(self):
        bio = "An amazing band"
        self._test_add_column(
            column=Text(default=bio),
            column_name="bio",
            expected_value=bio,
        )

    def test_problematic_name(self):
        """
        Make sure we can add columns with names which clash with SQL keywords.
        """
        self._test_add_column(
            column=Text(default="asc"),
            column_name="order",
            expected_value="asc",
        )


class TestUnique(DBTestCase):
    @engines_only("postgres")
    def test_unique(self):
        """
        Test altering a column uniqueness with MigrationManager.
        🐛 Cockroach bug: https://github.com/cockroachdb/cockroach/issues/42840 "unimplemented: cannot drop UNIQUE constraint "manager_name_key" using ALTER TABLE DROP CONSTRAINT, use DROP INDEX CASCADE instead"
        """  # noqa: E501
        unique_query = Manager.alter().set_unique(Manager.name, True)
        unique_query.run_sync()

        Manager(name="Bob").save().run_sync()

        # Make sure non-unique names work:
        Manager(name="Sally").save().run_sync()

        # Check there's a unique row error ...
        with self.assertRaises(Exception):
            Manager(name="Bob").save().run_sync()

        response = Manager.select().run_sync()
        self.assertEqual(len(response), 2)

        # Now remove the constraint, and add a row.
        not_unique_query = Manager.alter().set_unique(Manager.name, False)
        not_unique_query.run_sync()
        Manager(name="Bob").save().run_sync()

        response = Manager.select().run_sync()
        self.assertTrue(len(response), 2)


@engines_only("postgres", "cockroach")
class TestMultiple(DBTestCase):
    """
    Make sure multiple alter statements work correctly.
    """

    def test_multiple(self):
        self.insert_row()

        query = (
            Manager.alter()
            .add_column("column_a", Integer(default=0, null=True))
            .add_column("column_b", Integer(default=0, null=True))
        )
        query.run_sync()

        response = Band.raw("SELECT * FROM manager").run_sync()

        column_names = response[0].keys()
        self.assertIn("column_a", column_names)
        self.assertIn("column_b", column_names)


# TODO - test more conversions.
@engines_only("postgres", "cockroach")
class TestSetColumnType(DBTestCase):
    def test_integer_to_bigint(self):
        """
        Test converting an Integer column to BigInt.
        """
        self.insert_row()

        alter_query = Band.alter().set_column_type(
            old_column=Band.popularity, new_column=BigInt()
        )
        alter_query.run_sync()

        self.assertEqual(
            self.get_postgres_column_type(
                tablename="band", column_name="popularity"
            ),
            "BIGINT",
        )

        row = Band.select(Band.popularity).first().run_sync()
        assert row is not None
        self.assertEqual(row["popularity"], 1000)

    def test_integer_to_varchar(self):
        """
        Test converting an Integer column to Varchar.
        """
        self.insert_row()

        alter_query = Band.alter().set_column_type(
            old_column=Band.popularity, new_column=Varchar()
        )
        alter_query.run_sync()

        self.assertEqual(
            self.get_postgres_column_type(
                tablename="band", column_name="popularity"
            ),
            "CHARACTER VARYING",
        )

        row = Band.select(Band.popularity).first().run_sync()
        assert row is not None
        self.assertEqual(row["popularity"], "1000")

    def test_using_expression(self):
        """
        Test the `using_expression` option, which can be used to tell Postgres
        how to convert certain column types.
        """
        Band(name="1").save().run_sync()

        alter_query = Band.alter().set_column_type(
            old_column=Band.name,
            new_column=Integer(),
            using_expression="name::integer",
        )
        alter_query.run_sync()

        row = Band.select(Band.name).first().run_sync()
        assert row is not None
        self.assertEqual(row["name"], 1)


@engines_only("postgres", "cockroach")
class TestSetNull(DBTestCase):
    def test_set_null(self):
        query = """
            SELECT is_nullable FROM information_schema.columns
            WHERE table_name = 'band'
            AND table_catalog = 'piccolo'
            AND column_name = 'popularity'
            """

        Band.alter().set_null(Band.popularity, boolean=True).run_sync()
        response = Band.raw(query).run_sync()
        self.assertEqual(response[0]["is_nullable"], "YES")

        Band.alter().set_null(Band.popularity, boolean=False).run_sync()
        response = Band.raw(query).run_sync()
        self.assertEqual(response[0]["is_nullable"], "NO")


@engines_only("postgres", "cockroach")
class TestSetLength(DBTestCase):
    def test_set_length(self):
        query = """
            SELECT character_maximum_length FROM information_schema.columns
            WHERE table_name = 'band'
            AND table_catalog = 'piccolo'
            AND column_name = 'name'
            """

        for length in (5, 20, 50):
            Band.alter().set_length(Band.name, length=length).run_sync()
            response = Band.raw(query).run_sync()
            self.assertEqual(response[0]["character_maximum_length"], length)


@engines_only("postgres", "cockroach")
class TestSetDefault(DBTestCase):
    def test_set_default(self):
        Manager.alter().set_default(Manager.name, "Pending").run_sync()

        # Bypassing the ORM to make sure the database default is present.
        Band.raw(
            "INSERT INTO manager (id, name) VALUES (DEFAULT, DEFAULT);"
        ).run_sync()

        manager = Manager.objects().first().run_sync()
        assert manager is not None
        self.assertEqual(manager.name, "Pending")


@engines_only("postgres", "cockroach")
class TestSetSchema(TestCase):
    schema_manager = SchemaManager()
    schema_name = "schema_1"

    def setUp(self):
        Manager.create_table().run_sync()
        self.schema_manager.create_schema(
            schema_name=self.schema_name
        ).run_sync()

    def tearDown(self):
        Manager.alter().drop_table(if_exists=True).run_sync()
        self.schema_manager.drop_schema(
            schema_name=self.schema_name, cascade=True
        ).run_sync()

    def test_set_schema(self):
        Manager.alter().set_schema(schema_name=self.schema_name).run_sync()

        self.assertIn(
            Manager._meta.tablename,
            self.schema_manager.list_tables(
                schema_name=self.schema_name
            ).run_sync(),
        )


@engines_only("postgres", "cockroach")
class TestDropTable(TestCase):
    class Manager(Table, schema="schema_1"):
        pass

    schema_manager = SchemaManager()

    def tearDown(self):
        self.schema_manager.drop_schema(
            schema_name="schema_1", if_exists=True, cascade=True
        ).run_sync()

    def test_drop_table_with_schema(self):
        Manager = self.Manager

        Manager.create_table().run_sync()

        self.assertIn(
            "manager",
            self.schema_manager.list_tables(schema_name="schema_1").run_sync(),
        )

        Manager.alter().drop_table().run_sync()

        self.assertNotIn(
            "manager",
            self.schema_manager.list_tables(schema_name="schema_1").run_sync(),
        )


###############################################################################


class Ticket(Table):
    price = Numeric(digits=(5, 2))


class TestSetDigits(TestCase):
    def setUp(self):
        Ticket.create_table().run_sync()

    def tearDown(self):
        Ticket.alter().drop_table().run_sync()

    @engines_only("postgres")
    def test_set_digits(self):
        query = """
            SELECT numeric_precision, numeric_scale
            FROM information_schema.columns
            WHERE table_name = 'ticket'
            AND table_catalog = 'piccolo'
            AND column_name = 'price'
            """

        Ticket.alter().set_digits(
            column=Ticket.price, digits=(6, 2)
        ).run_sync()
        response = Ticket.raw(query).run_sync()
        self.assertEqual(response[0]["numeric_precision"], 6)
        self.assertEqual(response[0]["numeric_scale"], 2)

        Ticket.alter().set_digits(column=Ticket.price, digits=None).run_sync()
        response = Ticket.raw(query).run_sync()
        self.assertIsNone(response[0]["numeric_precision"])
        self.assertIsNone(response[0]["numeric_scale"])



from .alter import Alter
from .count import Count
from .create import Create
from .create_index import CreateIndex
from .delete import Delete
from .drop_index import DropIndex
from .exists import Exists
from .insert import Insert
from .objects import Objects
from .raw import Raw
from .refresh import Refresh
from .select import Select
from .table_exists import TableExists
from .update import Update

__all__ = (
    "Alter",
    "Count",
    "Create",
    "CreateIndex",
    "Delete",
    "DropIndex",
    "Exists",
    "Insert",
    "Objects",
    "Raw",
    "Refresh",
    "Select",
    "TableExists",
    "Update",
)



from __future__ import annotations

import typing as t

from piccolo.utils.lazy_loader import LazyLoader
from piccolo.utils.warnings import Level, colored_warning

from .postgres import PostgresEngine

asyncpg = LazyLoader("asyncpg", globals(), "asyncpg")


class CockroachEngine(PostgresEngine):
    """
    An extension of
    :class:`PostgresEngine <piccolo.engine.postgres.PostgresEngine>`.
    """

    engine_type = "cockroach"
    min_version_number = 0  # Doesn't seem to work with cockroach versioning.

    def __init__(
        self,
        config: t.Dict[str, t.Any],
        extensions: t.Sequence[str] = (),
        log_queries: bool = False,
        log_responses: bool = False,
        extra_nodes: t.Optional[t.Dict[str, CockroachEngine]] = None,
    ) -> None:
        super().__init__(
            config=config,
            extensions=extensions,
            log_queries=log_queries,
            log_responses=log_responses,
            extra_nodes=extra_nodes,
        )

    async def prep_database(self):
        try:
            await self._run_in_new_connection(
                "SET CLUSTER SETTING sql.defaults.experimental_alter_column_type.enabled = true;"  # noqa: E501
            )
        except asyncpg.exceptions.InsufficientPrivilegeError:
            colored_warning(
                "=> Unable to set up Cockroach DB "
                "functionality may not behave as expected. Make sure "
                "your database user has permission to set cluster options.",
                level=Level.medium,
            )



from .aggregate import Avg, Count, Max, Min, Sum
from .string import Length, Lower, Ltrim, Reverse, Rtrim, Upper

__all__ = (
    "Avg",
    "Count",
    "Length",
    "Lower",
    "Ltrim",
    "Max",
    "Min",
    "Reverse",
    "Rtrim",
    "Sum",
    "Upper",
)



from __future__ import annotations

import pathlib
from unittest import TestCase

from piccolo.apps.user.tables import BaseUser
from piccolo.conf.apps import AppConfig, AppRegistry, Finder, table_finder
from tests.example_apps.mega.tables import MegaTable, SmallTable
from tests.example_apps.music.tables import (
    Band,
    Concert,
    Instrument,
    Manager,
    Poster,
    RecordingStudio,
    Shirt,
    Ticket,
    Venue,
)


class TestAppRegistry(TestCase):
    def test_get_app_config(self):
        app_registry = AppRegistry(apps=["piccolo.apps.user.piccolo_app"])
        app_config = app_registry.get_app_config(app_name="user")
        self.assertIsInstance(app_config, AppConfig)

    def test_get_table_classes(self):
        app_registry = AppRegistry(apps=["piccolo.apps.user.piccolo_app"])
        table_classes = app_registry.get_table_classes(app_name="user")
        self.assertIn(BaseUser, table_classes)

        with self.assertRaises(ValueError):
            app_registry.get_table_classes(app_name="Foo")

    def test_duplicate_app_names(self):
        """
        An exception should be if apps with duplicate names are registered.
        """
        with self.assertRaises(ValueError):
            AppRegistry(
                apps=[
                    "piccolo.apps.user.piccolo_app",
                    "piccolo.apps.user.piccolo_app",
                ]
            )

    def test_app_names_not_ending_piccolo_app(self):
        """
        Should automatically add `.piccolo_app` to end.
        """
        AppRegistry(
            apps=[
                "piccolo.apps.user",
            ]
        )

    def test_duplicate_app_names_with_auto_changed(self):
        """
        Make sure duplicate app names are still detected when `piccolo_app`
        is omitted from the end.
        """
        with self.assertRaises(ValueError):
            AppRegistry(
                apps=[
                    "piccolo.apps.user.piccolo_app",
                    "piccolo.apps.user",
                ]
            )

    def test_get_table_with_name(self):
        app_registry = AppRegistry(apps=["piccolo.apps.user.piccolo_app"])
        table = app_registry.get_table_with_name(
            app_name="user", table_class_name="BaseUser"
        )
        self.assertEqual(table, BaseUser)


class TestAppConfig(TestCase):
    def test_pathlib(self):
        """
        Make sure a ``pathlib.Path`` instance can be passed in as a
        ``migrations_folder_path`` argument.
        """
        config = AppConfig(
            app_name="music", migrations_folder_path=pathlib.Path(__file__)
        )
        self.assertEqual(config.migrations_folder_path, __file__)

    def test_get_table_with_name(self):
        """
        Register a table, then test retrieving it.
        """
        config = AppConfig(app_name="music", migrations_folder_path="")
        config.register_table(table_class=Manager)
        self.assertEqual(config.get_table_with_name("Manager"), Manager)

        with self.assertRaises(ValueError):
            config.get_table_with_name("Foo")


class TestTableFinder(TestCase):
    def test_table_finder(self):
        """
        Should return all Table subclasses.
        """
        tables = table_finder(modules=["tests.example_apps.music.tables"])

        table_class_names = [i.__name__ for i in tables]
        table_class_names.sort()

        self.assertEqual(
            table_class_names,
            [
                "Band",
                "Concert",
                "Instrument",
                "Manager",
                "Poster",
                "RecordingStudio",
                "Shirt",
                "Ticket",
                "Venue",
            ],
        )

        with self.assertRaises(ImportError):
            table_finder(modules=["foo.bar.baz"])

    def test_table_finder_coercion(self):
        """
        Should convert a string argument to a list.
        """
        tables = table_finder(modules="tests.example_apps.music.tables")

        table_class_names = [i.__name__ for i in tables]
        table_class_names.sort()

        self.assertEqual(
            table_class_names,
            [
                "Band",
                "Concert",
                "Instrument",
                "Manager",
                "Poster",
                "RecordingStudio",
                "Shirt",
                "Ticket",
                "Venue",
            ],
        )

    def test_include_tags(self):
        """
        Should return all Table subclasses with a matching tag.
        """
        tables = table_finder(
            modules=["tests.example_apps.music.tables"],
            include_tags=["special"],
        )

        table_class_names = [i.__name__ for i in tables]
        table_class_names.sort()

        self.assertEqual(
            table_class_names,
            ["Poster"],
        )

    def test_exclude_tags(self):
        """
        Should return all Table subclasses without the specified tags.
        """
        tables = table_finder(
            modules=["tests.example_apps.music.tables"],
            exclude_tags=["special"],
        )

        table_class_names = [i.__name__ for i in tables]
        table_class_names.sort()

        self.assertEqual(
            table_class_names,
            [
                "Band",
                "Concert",
                "Instrument",
                "Manager",
                "RecordingStudio",
                "Shirt",
                "Ticket",
                "Venue",
            ],
        )

    def test_exclude_imported(self):
        """
        Make sure we can excluded imported Tables.
        """
        filtered_tables = table_finder(
            modules=["tests.conf.example"],
            exclude_imported=True,
        )

        self.assertEqual(
            [i.__name__ for i in filtered_tables],
            ["Musician"],
        )

        # Now try without filtering:
        all_tables = table_finder(
            modules=["tests.conf.example"],
            exclude_imported=False,
        )

        self.assertEqual(
            sorted([i.__name__ for i in all_tables]),
            ["BaseUser", "Musician"],
        )


class TestFinder(TestCase):
    def test_get_table_classes(self):
        """
        Make sure ``Table`` classes can be retrieved.
        """
        finder = Finder()

        self.assertListEqual(
            sorted(finder.get_table_classes(), key=lambda i: i.__name__),
            [
                Band,
                Concert,
                Instrument,
                Manager,
                MegaTable,
                Poster,
                RecordingStudio,
                Shirt,
                SmallTable,
                Ticket,
                Venue,
            ],
        )

        self.assertListEqual(
            sorted(
                finder.get_table_classes(include_apps=["music"]),
                key=lambda i: i.__name__,
            ),
            [
                Band,
                Concert,
                Instrument,
                Manager,
                Poster,
                RecordingStudio,
                Shirt,
                Ticket,
                Venue,
            ],
        )

        self.assertListEqual(
            sorted(
                finder.get_table_classes(exclude_apps=["music"]),
                key=lambda i: i.__name__,
            ),
            [
                MegaTable,
                SmallTable,
            ],
        )

        with self.assertRaises(ValueError):
            # You shouldn't be allowed to specify both include and exclude.
            finder.get_table_classes(
                exclude_apps=["music"], include_apps=["mega"]
            )

    def test_sort_app_configs(self):
        """
        Make sure we can sort ``AppConfig`` based on their migration
        dependencies.
        """
        app_config_1 = AppConfig(
            app_name="app_1",
            migrations_folder_path="",
        )

        app_config_1._migration_dependency_app_configs = [
            AppConfig(
                app_name="app_2",
                migrations_folder_path="",
            )
        ]

        app_config_2 = AppConfig(
            app_name="app_2",
            migrations_folder_path="",
        )

        app_config_2._migration_dependency_app_configs = []

        sorted_app_configs = Finder().sort_app_configs(
            [app_config_2, app_config_1]
        )

        self.assertListEqual(
            [i.app_name for i in sorted_app_configs], ["app_2", "app_1"]
        )



from unittest import TestCase

from tests.base import engines_skip

from .base import M2MBase


@engines_skip("sqlite")
class TestM2MWithSchema(M2MBase, TestCase):
    """
    Make sure that when the tables exist in a non-public schema, that M2M still
    works.
    """

    def setUp(self):
        return self._setUp(schema="schema_1")



from __future__ import annotations

import itertools
import typing as t
from collections import OrderedDict

from piccolo.columns import Column, Selectable
from piccolo.columns.column_types import JSON, JSONB, PrimaryKey
from piccolo.columns.m2m import M2MSelect
from piccolo.columns.readable import Readable
from piccolo.custom_types import TableInstance
from piccolo.engine.base import Batch
from piccolo.query.base import Query
from piccolo.query.mixins import (
    AsOfDelegate,
    CallbackDelegate,
    CallbackType,
    ColumnsDelegate,
    DistinctDelegate,
    GroupByDelegate,
    LimitDelegate,
    OffsetDelegate,
    OrderByDelegate,
    OrderByRaw,
    OutputDelegate,
    WhereDelegate,
)
from piccolo.query.proxy import Proxy
from piccolo.querystring import QueryString
from piccolo.utils.dictionary import make_nested
from piccolo.utils.encoding import dump_json, load_json
from piccolo.utils.warnings import colored_warning

if t.TYPE_CHECKING:  # pragma: no cover
    from piccolo.custom_types import Combinable
    from piccolo.table import Table  # noqa

# Here to avoid breaking changes - will be removed in the future.
from piccolo.query.functions.aggregate import Count  # noqa: F401


class SelectRaw(Selectable):
    def __init__(self, sql: str, *args: t.Any) -> None:
        """
        Execute raw SQL in your select query.

        .. code-block:: python

            >>> await Band.select(
            ...     Band.name,
            ...     SelectRaw("log(popularity) AS log_popularity")
            ... )
            [{'name': 'Pythonistas', 'log_popularity': 3.0}]

        """
        self.querystring = QueryString(sql, *args)

    def get_select_string(
        self, engine_type: str, with_alias: bool = True
    ) -> QueryString:
        return self.querystring


OptionalDict = t.Optional[t.Dict[str, t.Any]]


class First(Proxy["Select", OptionalDict]):
    """
    This is for static typing purposes.
    """

    def __init__(self, query: Select):
        self.query = query

    async def run(
        self,
        node: t.Optional[str] = None,
        in_pool: bool = True,
    ) -> OptionalDict:
        rows = await self.query.run(
            node=node, in_pool=in_pool, use_callbacks=False
        )
        results = rows[0] if rows else None

        modified_response = await self.query.callback_delegate.invoke(
            results=results, kind=CallbackType.success
        )
        return modified_response


class SelectList(Proxy["Select", t.List]):
    """
    This is for static typing purposes.
    """

    async def run(
        self,
        node: t.Optional[str] = None,
        in_pool: bool = True,
    ) -> t.List:
        rows = await self.query.run(
            node=node, in_pool=in_pool, use_callbacks=False
        )

        if len(rows) == 0:
            response = []
        else:
            if len(rows[0].keys()) != 1:
                raise ValueError("Each row returned more than one value")

            response = list(itertools.chain(*[j.values() for j in rows]))

        modified_response = await self.query.callback_delegate.invoke(
            results=response, kind=CallbackType.success
        )
        return modified_response


class SelectJSON(Proxy["Select", str]):
    """
    This is for static typing purposes.
    """

    async def run(
        self,
        node: t.Optional[str] = None,
        in_pool: bool = True,
    ) -> str:
        rows = await self.query.run(node=node, in_pool=in_pool)
        return dump_json(rows)


class Select(Query[TableInstance, t.List[t.Dict[str, t.Any]]]):
    __slots__ = (
        "columns_list",
        "exclude_secrets",
        "as_of_delegate",
        "columns_delegate",
        "distinct_delegate",
        "group_by_delegate",
        "limit_delegate",
        "offset_delegate",
        "order_by_delegate",
        "output_delegate",
        "callback_delegate",
        "where_delegate",
    )

    def __init__(
        self,
        table: t.Type[TableInstance],
        columns_list: t.Optional[t.Sequence[t.Union[Selectable, str]]] = None,
        exclude_secrets: bool = False,
        **kwargs,
    ):
        if columns_list is None:
            columns_list = []
        super().__init__(table, **kwargs)
        self.exclude_secrets = exclude_secrets

        self.as_of_delegate = AsOfDelegate()
        self.columns_delegate = ColumnsDelegate()
        self.distinct_delegate = DistinctDelegate()
        self.group_by_delegate = GroupByDelegate()
        self.limit_delegate = LimitDelegate()
        self.offset_delegate = OffsetDelegate()
        self.order_by_delegate = OrderByDelegate()
        self.output_delegate = OutputDelegate()
        self.callback_delegate = CallbackDelegate()
        self.where_delegate = WhereDelegate()

        self.columns(*columns_list)

    def columns(self: Self, *columns: t.Union[Selectable, str]) -> Self:
        _columns = self.table._process_column_args(*columns)
        self.columns_delegate.columns(*_columns)
        return self

    def distinct(
        self: Self, *, on: t.Optional[t.Sequence[Column]] = None
    ) -> Self:
        if on is not None and self.engine_type == "sqlite":
            raise NotImplementedError("SQLite doesn't support DISTINCT ON")

        self.distinct_delegate.distinct(enabled=True, on=on)
        return self

    def group_by(self: Self, *columns: t.Union[Column, str]) -> Self:
        _columns: t.List[Column] = [
            i
            for i in self.table._process_column_args(*columns)
            if isinstance(i, Column)
        ]
        self.group_by_delegate.group_by(*_columns)
        return self

    def as_of(self: Self, interval: str = "-1s") -> Self:
        if self.engine_type != "cockroach":
            raise NotImplementedError("Only CockroachDB supports AS OF")

        self.as_of_delegate.as_of(interval)
        return self

    def limit(self: Self, number: int) -> Self:
        self.limit_delegate.limit(number)
        return self

    def first(self) -> First:
        self.limit_delegate.limit(1)
        return First(query=self)

    def offset(self: Self, number: int) -> Self:
        self.offset_delegate.offset(number)
        return self

    async def _splice_m2m_rows(
        self,
        response: t.List[t.Dict[str, t.Any]],
        secondary_table: t.Type[Table],
        secondary_table_pk: PrimaryKey,
        m2m_name: str,
        m2m_select: M2MSelect,
        as_list: bool = False,
    ):
        row_ids = list(
            set(itertools.chain(*[row[m2m_name] for row in response]))
        )
        extra_rows = (
            (
                await secondary_table.select(
                    *m2m_select.columns,
                    secondary_table_pk.as_alias("mapping_key"),
                )
                .where(secondary_table_pk.is_in(row_ids))
                .output(load_json=m2m_select.load_json)
                .run()
            )
            if row_ids
            else []
        )
        if as_list:
            column_name = m2m_select.columns[0]._meta.name
            extra_rows_map = {
                row["mapping_key"]: row[column_name] for row in extra_rows
            }
        else:
            extra_rows_map = {
                row["mapping_key"]: {
                    key: value
                    for key, value in row.items()
                    if key != "mapping_key"
                }
                for row in extra_rows
            }
        for row in response:
            row[m2m_name] = [extra_rows_map.get(i) for i in row[m2m_name]]
        return response

    async def response_handler(self, response):
        m2m_selects = [
            i
            for i in self.columns_delegate.selected_columns
            if isinstance(i, M2MSelect)
        ]
        for m2m_select in m2m_selects:
            m2m_name = m2m_select.m2m._meta.name
            secondary_table = m2m_select.m2m._meta.secondary_table
            secondary_table_pk = secondary_table._meta.primary_key

            if self.engine_type == "sqlite":
                # With M2M queries in SQLite, we always get the value back as a
                # list of strings, so we need to do some type conversion.
                value_type = (
                    m2m_select.columns[0].__class__.value_type
                    if m2m_select.as_list and m2m_select.serialisation_safe
                    else secondary_table_pk.value_type
                )
                try:
                    for row in response:
                        data = row[m2m_name]
                        row[m2m_name] = (
                            [value_type(i) for i in row[m2m_name]]
                            if data
                            else []
                        )
                except ValueError:
                    colored_warning(
                        "Unable to do type conversion for the "
                        f"{m2m_name} relation"
                    )

                # If the user requested a single column, we just return that
                # from the database. Otherwise we request the primary key
                # value, so we can fetch the rest of the data in a subsequent
                # SQL query - see below.
                if m2m_select.as_list:
                    if m2m_select.serialisation_safe:
                        pass
                    else:
                        response = await self._splice_m2m_rows(
                            response,
                            secondary_table,
                            secondary_table_pk,
                            m2m_name,
                            m2m_select,
                            as_list=True,
                        )
                else:
                    if (
                        len(m2m_select.columns) == 1
                        and m2m_select.serialisation_safe
                    ):
                        column_name = m2m_select.columns[0]._meta.name
                        for row in response:
                            row[m2m_name] = [
                                {column_name: i} for i in row[m2m_name]
                            ]
                    else:
                        response = await self._splice_m2m_rows(
                            response,
                            secondary_table,
                            secondary_table_pk,
                            m2m_name,
                            m2m_select,
                        )

            elif self.engine_type in ("postgres", "cockroach"):
                if m2m_select.as_list:
                    # We get the data back as an array, and can just return it
                    # unless it's JSON.
                    if (
                        type(m2m_select.columns[0]) in (JSON, JSONB)
                        and m2m_select.load_json
                    ):
                        for row in response:
                            data = row[m2m_name]
                            row[m2m_name] = [load_json(i) for i in data]
                elif m2m_select.serialisation_safe:
                    # If the columns requested can be safely serialised, they
                    # are returned as a JSON string, so we need to deserialise
                    # it.
                    for row in response:
                        data = row[m2m_name]
                        row[m2m_name] = load_json(data) if data else []
                else:
                    # If the data can't be safely serialised as JSON, we get
                    # back an array of primary key values, and need to
                    # splice in the correct values using Python.
                    response = await self._splice_m2m_rows(
                        response,
                        secondary_table,
                        secondary_table_pk,
                        m2m_name,
                        m2m_select,
                    )

        #######################################################################

        # If no columns were specified, it's a select *, so we know that
        # no columns were selected from related tables.
        was_select_star = len(self.columns_delegate.selected_columns) == 0

        if self.output_delegate._output.nested and not was_select_star:
            return [make_nested(i) for i in response]
        else:
            return response

    def order_by(
        self: Self, *columns: t.Union[Column, str, OrderByRaw], ascending=True
    ) -> Self:
        """
        :param columns:
            Either a :class:`piccolo.columns.base.Column` instance, a string
            representing a column name, or :class:`piccolo.query.OrderByRaw`
            which allows you for complex use cases like
            ``OrderByRaw('random()')``.
        """
        _columns: t.List[t.Union[Column, OrderByRaw]] = []
        for column in columns:
            if isinstance(column, str):
                _columns.append(self.table._meta.get_column_by_name(column))
            else:
                _columns.append(column)

        self.order_by_delegate.order_by(*_columns, ascending=ascending)
        return self

    @t.overload
    def output(self: Self, *, as_list: bool) -> SelectList: ...

    @t.overload
    def output(self: Self, *, as_json: bool) -> SelectJSON: ...

    @t.overload
    def output(self: Self, *, load_json: bool) -> Self: ...

    @t.overload
    def output(self: Self, *, nested: bool) -> Self: ...

    def output(
        self: Self,
        *,
        as_list: bool = False,
        as_json: bool = False,
        load_json: bool = False,
        nested: bool = False,
    ):
        self.output_delegate.output(
            as_list=as_list,
            as_json=as_json,
            load_json=load_json,
            nested=nested,
        )
        if as_list:
            return SelectList(query=self)
        elif as_json:
            return SelectJSON(query=self)

        return self

    def callback(
        self: Self,
        callbacks: t.Union[t.Callable, t.List[t.Callable]],
        *,
        on: CallbackType = CallbackType.success,
    ) -> Self:
        self.callback_delegate.callback(callbacks, on=on)
        return self

    def where(self: Self, *where: t.Union[Combinable, QueryString]) -> Self:
        self.where_delegate.where(*where)
        return self

    async def batch(
        self,
        batch_size: t.Optional[int] = None,
        node: t.Optional[str] = None,
        **kwargs,
    ) -> Batch:
        if batch_size:
            kwargs.update(batch_size=batch_size)
        if node:
            kwargs.update(node=node)
        return await self.table._meta.db.batch(self, **kwargs)

    ###########################################################################

    def _get_joins(self, columns: t.Sequence[Selectable]) -> t.List[str]:
        """
        A call chain is a sequence of foreign keys representing joins which
        need to be made to retrieve a column in another table.
        """
        joins: t.List[str] = []

        readables: t.List[Readable] = [
            i for i in columns if isinstance(i, Readable)
        ]

        columns = list(columns)
        for readable in readables:
            columns += readable.columns

        querystrings: t.List[QueryString] = [
            i for i in columns if isinstance(i, QueryString)
        ]
        for querystring in querystrings:
            if querystring_columns := getattr(querystring, "columns", []):
                columns += querystring_columns

        for column in columns:
            if not isinstance(column, Column):
                continue

            _joins: t.List[str] = []
            for index, key in enumerate(column._meta.call_chain, 0):
                table_alias = key.table_alias

                if index > 0:
                    left_tablename = column._meta.call_chain[
                        index - 1
                    ].table_alias
                else:
                    left_tablename = (
                        key._meta.table._meta.get_formatted_tablename()
                    )  # noqa: E501

                right_tablename = (
                    key._foreign_key_meta.resolved_references._meta.get_formatted_tablename()  # noqa: E501
                )

                pk_name = column._meta.call_chain[
                    index
                ]._foreign_key_meta.resolved_target_column._meta.name

                _joins.append(
                    f'LEFT JOIN {right_tablename} "{table_alias}"'
                    " ON "
                    f'({left_tablename}."{key._meta.name}" = "{table_alias}"."{pk_name}")'  # noqa: E501
                )

            joins.extend(_joins)

        # Remove duplicates
        return list(OrderedDict.fromkeys(joins))

    def _check_valid_call_chain(self, keys: t.Sequence[Selectable]) -> bool:
        for column in keys:
            if not isinstance(column, Column):
                continue
            if column._meta.call_chain and len(column._meta.call_chain) > 10:
                # Make sure the call_chain isn't too large to discourage
                # very inefficient queries.
                raise Exception(
                    "Joining more than 10 tables isn't supported - "
                    "please restructure your query."
                )
        return True

    @property
    def default_querystrings(self) -> t.Sequence[QueryString]:
        # JOIN
        self._check_valid_call_chain(self.columns_delegate.selected_columns)

        select_joins = self._get_joins(self.columns_delegate.selected_columns)
        where_joins = self._get_joins(self.where_delegate.get_where_columns())
        order_by_joins = self._get_joins(
            self.order_by_delegate.get_order_by_columns()
        )

        # Combine all joins, and remove duplicates
        joins: t.List[str] = list(
            OrderedDict.fromkeys(select_joins + where_joins + order_by_joins)
        )

        #######################################################################

        # If no columns have been specified for selection, select all columns
        # on the table:
        if len(self.columns_delegate.selected_columns) == 0:
            self.columns_delegate.selected_columns = self.table._meta.columns

        # If secret fields need to be omitted, remove them from the list.
        if self.exclude_secrets:
            self.columns_delegate.remove_secret_columns()

        engine_type = self.table._meta.db.engine_type

        select_strings: t.List[QueryString] = [
            c.get_select_string(engine_type=engine_type)
            for c in self.columns_delegate.selected_columns
        ]

        #######################################################################

        args: t.List[t.Any] = []

        query = "SELECT"

        distinct = self.distinct_delegate._distinct
        if distinct.on:
            distinct.validate_on(self.order_by_delegate._order_by)
        query += "{}"
        args.append(distinct.querystring)

        columns_str = ", ".join("{}" for i in select_strings)
        query += f" {columns_str} FROM {self.table._meta.get_formatted_tablename()}"  # noqa: E501
        args.extend(select_strings)

        for join in joins:
            query += f" {join}"

        if self.as_of_delegate._as_of:
            query += "{}"
            args.append(self.as_of_delegate._as_of.querystring)

        if self.where_delegate._where:
            query += " WHERE {}"
            args.append(self.where_delegate._where.querystring)

        if self.group_by_delegate._group_by:
            query += "{}"
            args.append(self.group_by_delegate._group_by.querystring)

        if self.order_by_delegate._order_by.order_by_items:
            query += "{}"
            args.append(self.order_by_delegate._order_by.querystring)

        if (
            engine_type == "sqlite"
            and self.offset_delegate._offset
            and not self.limit_delegate._limit
        ):
            raise ValueError(
                "A limit clause must be provided when doing an offset with "
                "SQLite."
            )

        if self.limit_delegate._limit:
            query += "{}"
            args.append(self.limit_delegate._limit.querystring)

        if self.offset_delegate._offset:
            query += "{}"
            args.append(self.offset_delegate._offset.querystring)

        querystring = QueryString(query, *args)

        return [querystring]

    async def run(
        self,
        node: t.Optional[str] = None,
        in_pool: bool = True,
        use_callbacks: bool = True,
        **kwargs,
    ) -> t.List[t.Dict[str, t.Any]]:
        results = await super().run(node=node, in_pool=in_pool)
        if use_callbacks:
            return await self.callback_delegate.invoke(
                results, kind=CallbackType.success
            )
        else:
            return results


Self = t.TypeVar("Self", bound=Select)



import datetime
from unittest import TestCase

from piccolo.columns.column_types import Date
from piccolo.columns.defaults.date import DateNow
from piccolo.table import Table


class MyTable(Table):
    created_on = Date()


class MyTableDefault(Table):
    created_on = Date(default=DateNow())


class TestDate(TestCase):
    def setUp(self):
        MyTable.create_table().run_sync()

    def tearDown(self):
        MyTable.alter().drop_table().run_sync()

    def test_timestamp(self):
        created_on = datetime.datetime.now().date()
        row = MyTable(created_on=created_on)
        row.save().run_sync()

        result = MyTable.objects().first().run_sync()
        assert result is not None
        self.assertEqual(result.created_on, created_on)


class TestDateDefault(TestCase):
    def setUp(self):
        MyTableDefault.create_table().run_sync()

    def tearDown(self):
        MyTableDefault.alter().drop_table().run_sync()

    def test_timestamp(self):
        created_on = datetime.datetime.now().date()
        row = MyTableDefault()
        row.save().run_sync()

        result = MyTableDefault.objects().first().run_sync()
        assert result is not None
        self.assertEqual(result.created_on, created_on)



import asyncio
import random
import typing as t
from io import StringIO
from unittest import TestCase
from unittest.mock import MagicMock, patch

from piccolo.apps.migrations.auto.migration_manager import MigrationManager
from piccolo.apps.migrations.commands.base import BaseMigrationManager
from piccolo.columns import Text, Varchar
from piccolo.columns.base import OnDelete, OnUpdate
from piccolo.columns.column_types import ForeignKey
from piccolo.conf.apps import AppConfig
from piccolo.table import Table, sort_table_classes
from piccolo.utils.lazy_loader import LazyLoader
from tests.base import AsyncMock, DBTestCase, engine_is, engines_only
from tests.example_apps.music.tables import Band, Concert, Manager, Venue

asyncpg = LazyLoader("asyncpg", globals(), "asyncpg")


class TestSortTableClasses(TestCase):
    def test_sort_table_classes(self):
        """
        Make sure simple use cases work correctly.
        """
        self.assertListEqual(
            sort_table_classes([Manager, Band]), [Manager, Band]
        )
        self.assertListEqual(
            sort_table_classes([Band, Manager]), [Manager, Band]
        )

        sorted_tables = sort_table_classes([Manager, Venue, Concert, Band])
        self.assertTrue(
            sorted_tables.index(Manager) < sorted_tables.index(Band)
        )
        self.assertTrue(
            sorted_tables.index(Venue) < sorted_tables.index(Concert)
        )
        self.assertTrue(
            sorted_tables.index(Band) < sorted_tables.index(Concert)
        )

    def test_sort_unrelated_tables(self):
        """
        Make sure there are no weird edge cases with tables with no foreign
        key relationships with each other.
        """

        class TableA(Table):
            pass

        class TableB(Table):
            pass

        self.assertListEqual(
            sort_table_classes([TableA, TableB]), [TableA, TableB]
        )

    def test_single_table(self):
        """
        Make sure that sorting a list with only a single table in it still
        works.
        """
        self.assertListEqual(sort_table_classes([Band]), [Band])

    def test_recursive_table(self):
        """
        Make sure that a table with a foreign key to itself sorts without
        issues.
        """

        class TableA(Table):
            table_a = ForeignKey("self")

        class TableB(Table):
            table_a = ForeignKey(TableA)

        self.assertListEqual(
            sort_table_classes([TableA, TableB]), [TableA, TableB]
        )

    def test_long_chain(self):
        """
        Make sure sorting works when there are a lot of tables with foreign
        keys to each other.

        https://github.com/piccolo-orm/piccolo/issues/616

        """

        class TableA(Table):
            pass

        class TableB(Table):
            fk = ForeignKey(TableA)

        class TableC(Table):
            fk = ForeignKey(TableB)

        class TableD(Table):
            fk = ForeignKey(TableC)

        class TableE(Table):
            fk = ForeignKey(TableD)

        tables = [TableA, TableB, TableC, TableD, TableE]

        shuffled_tables = random.sample(tables, len(tables))

        self.assertListEqual(sort_table_classes(shuffled_tables), tables)


class TestMigrationManager(DBTestCase):
    @engines_only("postgres", "cockroach")
    def test_rename_column(self):
        """
        Test running a MigrationManager which contains a column rename
        operation.
        """
        self.insert_row()

        manager = MigrationManager()
        manager.rename_column(
            table_class_name="Band",
            tablename="band",
            old_column_name="name",
            new_column_name="title",
        )
        asyncio.run(manager.run())

        response = self.run_sync("SELECT * FROM band;")
        self.assertTrue("title" in response[0].keys())
        self.assertTrue("name" not in response[0].keys())

        # Reverse
        asyncio.run(manager.run(backwards=True))
        response = self.run_sync("SELECT * FROM band;")
        self.assertTrue("title" not in response[0].keys())
        self.assertTrue("name" in response[0].keys())

        # Preview
        manager.preview = True
        with patch("sys.stdout", new=StringIO()) as fake_out:
            asyncio.run(manager.run())
            self.assertEqual(
                fake_out.getvalue(),
                """  -  [preview forwards]... \n ALTER TABLE "band" RENAME COLUMN "name" TO "title";\n""",  # noqa: E501
            )
        response = self.run_sync("SELECT * FROM band;")
        self.assertTrue("title" not in response[0].keys())
        self.assertTrue("name" in response[0].keys())

    def test_raw_function(self):
        """
        Test adding raw functions to a MigrationManager.
        """

        class HasRun(Exception):
            pass

        class HasRunBackwards(Exception):
            pass

        def run():
            raise HasRun("I was run!")

        def run_back():
            raise HasRunBackwards("I was run backwards!")

        manager = MigrationManager()
        manager.add_raw(run)
        manager.add_raw_backwards(run_back)

        with self.assertRaises(HasRun):
            asyncio.run(manager.run())

        # Reverse
        with self.assertRaises(HasRunBackwards):
            asyncio.run(manager.run(backwards=True))

    def test_raw_coroutine(self):
        """
        Test adding raw coroutines to a MigrationManager.
        """

        class HasRun(Exception):
            pass

        class HasRunBackwards(Exception):
            pass

        async def run():
            raise HasRun("I was run!")

        async def run_back():
            raise HasRunBackwards("I was run backwards!")

        manager = MigrationManager()
        manager.add_raw(run)
        manager.add_raw_backwards(run_back)

        with self.assertRaises(HasRun):
            asyncio.run(manager.run())

        # Reverse
        with self.assertRaises(HasRunBackwards):
            asyncio.run(manager.run(backwards=True))

    @engines_only("postgres", "cockroach")
    @patch.object(BaseMigrationManager, "get_app_config")
    def test_add_table(self, get_app_config: MagicMock):
        """
        Test adding a table to a MigrationManager.
        """
        self.run_sync("DROP TABLE IF EXISTS musician;")

        manager = MigrationManager()
        manager.add_table(class_name="Musician", tablename="musician")
        manager.add_column(
            table_class_name="Musician",
            tablename="musician",
            column_name="name",
            column_class_name="Varchar",
        )
        asyncio.run(manager.run())

        if engine_is("postgres"):
            self.run_sync(
                "INSERT INTO musician VALUES (default, 'Bob Jones');"
            )
            response = self.run_sync("SELECT * FROM musician;")
            self.assertEqual(response, [{"id": 1, "name": "Bob Jones"}])

        if engine_is("cockroach"):
            id = self.run_sync(
                "INSERT INTO musician VALUES (default, 'Bob Jones') RETURNING id;"  # noqa: E501
            )
            response = self.run_sync("SELECT * FROM musician;")
            self.assertEqual(
                response, [{"id": id[0]["id"], "name": "Bob Jones"}]
            )

        # Reverse
        get_app_config.return_value = AppConfig(
            app_name="music", migrations_folder_path=""
        )
        asyncio.run(manager.run(backwards=True))
        self.assertEqual(self.table_exists("musician"), False)
        self.run_sync("DROP TABLE IF EXISTS musician;")

        # Preview
        manager.preview = True
        with patch("sys.stdout", new=StringIO()) as fake_out:
            asyncio.run(manager.run())

            if engine_is("postgres"):
                self.assertEqual(
                    fake_out.getvalue(),
                    """  -  [preview forwards]... \n CREATE TABLE "musician" ("id" SERIAL PRIMARY KEY NOT NULL, "name" VARCHAR(255) NOT NULL DEFAULT '');\n""",  # noqa: E501
                )
            if engine_is("cockroach"):
                self.assertEqual(
                    fake_out.getvalue(),
                    """  -  [preview forwards]... \n CREATE TABLE "musician" ("id" INTEGER PRIMARY KEY NOT NULL DEFAULT unique_rowid(), "name" VARCHAR(255) NOT NULL DEFAULT '');\n""",  # noqa: E501
                )
        self.assertEqual(self.table_exists("musician"), False)

    @engines_only("postgres", "cockroach")
    def test_add_column(self) -> None:
        """
        Test adding a column to a MigrationManager.
        """
        manager = MigrationManager()
        manager.add_column(
            table_class_name="Manager",
            tablename="manager",
            column_name="email",
            column_class=Varchar,
            column_class_name="Varchar",
            params={
                "length": 100,
                "default": "",
                "null": True,
                "primary": False,
                "key": False,
                "unique": True,
                "index": False,
            },
        )
        asyncio.run(manager.run())

        if engine_is("postgres"):
            self.run_sync(
                "INSERT INTO \"manager\" VALUES (default, 'Dave', 'dave@me.com');"  # noqa: E501
            )
            response = self.run_sync("SELECT * FROM manager;")
            self.assertEqual(
                response, [{"id": 1, "name": "Dave", "email": "dave@me.com"}]
            )

            # Reverse
            asyncio.run(manager.run(backwards=True))
            response = self.run_sync("SELECT * FROM manager;")
            self.assertEqual(response, [{"id": 1, "name": "Dave"}])

        row_id: t.Optional[int] = None
        if engine_is("cockroach"):
            row_id = self.run_sync(
                "INSERT INTO manager VALUES (default, 'Dave', 'dave@me.com') RETURNING id;"  # noqa: E501
            )[0]["id"]
            response = self.run_sync("SELECT * FROM manager;")
            self.assertEqual(
                response,
                [{"id": row_id, "name": "Dave", "email": "dave@me.com"}],
            )

            # Reverse
            asyncio.run(manager.run(backwards=True))
            response = self.run_sync("SELECT * FROM manager;")
            self.assertEqual(response, [{"id": row_id, "name": "Dave"}])

        # Preview
        manager.preview = True
        with patch("sys.stdout", new=StringIO()) as fake_out:
            asyncio.run(manager.run())
            self.assertEqual(
                fake_out.getvalue(),
                """  -  [preview forwards]... \n ALTER TABLE "manager" ADD COLUMN "email" VARCHAR(100) UNIQUE DEFAULT '';\n""",  # noqa: E501
            )

        response = self.run_sync("SELECT * FROM manager;")
        if engine_is("postgres"):
            self.assertEqual(response, [{"id": 1, "name": "Dave"}])
        if engine_is("cockroach"):
            self.assertEqual(response, [{"id": row_id, "name": "Dave"}])

    @engines_only("postgres", "cockroach")
    def test_add_column_with_index(self):
        """
        Test adding a column with an index to a MigrationManager.
        """
        manager = MigrationManager()
        manager.add_column(
            table_class_name="Manager",
            tablename="manager",
            column_name="email",
            column_class=Varchar,
            column_class_name="Varchar",
            params={
                "length": 100,
                "default": "",
                "null": True,
                "primary": False,
                "key": False,
                "unique": True,
                "index": True,
            },
        )
        index_name = Manager._get_index_name(["email"])

        asyncio.run(manager.run())
        self.assertTrue(index_name in Manager.indexes().run_sync())

        # Reverse
        asyncio.run(manager.run(backwards=True))
        self.assertTrue(index_name not in Manager.indexes().run_sync())

        # Preview
        manager.preview = True
        with patch("sys.stdout", new=StringIO()) as fake_out:
            asyncio.run(manager.run())
            self.assertEqual(
                fake_out.getvalue(),
                (
                    """  -  [preview forwards]... \n ALTER TABLE "manager" ADD COLUMN "email" VARCHAR(100) UNIQUE DEFAULT '';\n"""  # noqa: E501
                    """\n CREATE INDEX manager_email ON "manager" USING btree ("email");\n"""  # noqa: E501
                ),
            )
        self.assertTrue(index_name not in Manager.indexes().run_sync())

    @engines_only("postgres")
    def test_add_foreign_key_self_column(self):
        """
        Test adding a ForeignKey column to a MigrationManager, with a
        references argument of 'self'.
        """
        manager = MigrationManager()
        manager.add_column(
            table_class_name="Manager",
            tablename="manager",
            column_name="advisor",
            column_class=ForeignKey,
            column_class_name="ForeignKey",
            params={
                "references": "self",
                "on_delete": OnDelete.cascade,
                "on_update": OnUpdate.cascade,
                "default": None,
                "null": True,
                "primary": False,
                "key": False,
                "unique": False,
                "index": False,
            },
        )
        asyncio.run(manager.run())

        self.run_sync("INSERT INTO manager VALUES (default, 'Alice', null);")
        self.run_sync("INSERT INTO manager VALUES (default, 'Dave', 1);")

        response = self.run_sync("SELECT * FROM manager;")
        self.assertEqual(
            response,
            [
                {"id": 1, "name": "Alice", "advisor": None},
                {"id": 2, "name": "Dave", "advisor": 1},
            ],
        )

        # Reverse
        asyncio.run(manager.run(backwards=True))
        response = self.run_sync("SELECT * FROM manager;")
        self.assertEqual(
            response,
            [{"id": 1, "name": "Alice"}, {"id": 2, "name": "Dave"}],
        )

    @engines_only("cockroach")
    def test_add_foreign_key_self_column_alt(self):
        """
        Test adding a ForeignKey column to a MigrationManager, with a
        references argument of 'self'.
        """
        manager = MigrationManager()
        manager.add_column(
            table_class_name="Manager",
            tablename="manager",
            column_name="advisor",
            column_class=ForeignKey,
            column_class_name="ForeignKey",
            params={
                "references": "self",
                "on_delete": OnDelete.cascade,
                "on_update": OnUpdate.cascade,
                "default": None,
                "null": True,
                "primary": False,
                "key": False,
                "unique": False,
                "index": False,
            },
        )
        asyncio.run(manager.run())

        id = self.run_sync(
            "INSERT INTO manager VALUES (default, 'Alice', null) RETURNING id;"
        )
        id2 = Manager.raw(
            "INSERT INTO manager VALUES (default, 'Dave', {}) RETURNING id;",
            id[0]["id"],
        ).run_sync()

        response = self.run_sync("SELECT * FROM manager;")
        self.assertEqual(
            response,
            [
                {"id": id[0]["id"], "name": "Alice", "advisor": None},
                {"id": id2[0]["id"], "name": "Dave", "advisor": id[0]["id"]},
            ],
        )

        # Reverse
        asyncio.run(manager.run(backwards=True))
        response = self.run_sync("SELECT * FROM manager;")
        self.assertEqual(
            response,
            [
                {"id": id[0]["id"], "name": "Alice"},
                {"id": id2[0]["id"], "name": "Dave"},
            ],
        )

    @engines_only("postgres", "cockroach")
    def test_add_non_nullable_column(self):
        """
        Test adding a non nullable column to a MigrationManager.

        Need to handle it gracefully if rows already exist.
        """
        self.run_sync("INSERT INTO manager VALUES (default, 'Dave');")

        manager = MigrationManager()
        manager.add_column(
            table_class_name="Manager",
            tablename="manager",
            column_name="email",
            column_class=Varchar,
            column_class_name="Varchar",
            params={
                "length": 100,
                "default": "",
                "null": False,
                "primary": False,
                "key": False,
                "unique": True,
                "index": False,
            },
        )
        asyncio.run(manager.run())

    @engines_only("postgres", "cockroach")
    @patch.object(
        BaseMigrationManager, "get_migration_managers", new_callable=AsyncMock
    )
    @patch.object(BaseMigrationManager, "get_app_config")
    def test_drop_column(
        self, get_app_config: MagicMock, get_migration_managers: MagicMock
    ):
        """
        Test dropping a column with MigrationManager.
        """
        manager_1 = MigrationManager()
        manager_1.add_table(class_name="Musician", tablename="musician")
        manager_1.add_column(
            table_class_name="Musician",
            tablename="musician",
            column_name="name",
            column_class_name="Varchar",
        )
        asyncio.run(manager_1.run())

        if engine_is("postgres"):
            self.run_sync("INSERT INTO musician VALUES (default, 'Dave');")
            response = self.run_sync("SELECT * FROM musician;")
            self.assertEqual(response, [{"id": 1, "name": "Dave"}])

        id = 0
        if engine_is("cockroach"):
            id = self.run_sync(
                "INSERT INTO musician VALUES (default, 'Dave') RETURNING id;"
            )
            response = self.run_sync("SELECT * FROM musician;")
            self.assertEqual(
                response, [{"id": id[0]["id"], "name": "Dave"}]  # type: ignore
            )

        manager_2 = MigrationManager()
        manager_2.drop_column(
            table_class_name="Musician",
            tablename="musician",
            column_name="name",
        )
        asyncio.run(manager_2.run())

        if engine_is("postgres"):
            response = self.run_sync("SELECT * FROM musician;")
            self.assertEqual(response, [{"id": 1}])

        if engine_is("cockroach"):
            response = self.run_sync("SELECT * FROM musician;")
            self.assertEqual(response, [{"id": id[0]["id"]}])  # type: ignore

        # Reverse
        get_migration_managers.return_value = [manager_1]
        app_config = AppConfig(app_name="music", migrations_folder_path="")
        get_app_config.return_value = app_config
        asyncio.run(manager_2.run(backwards=True))
        response = self.run_sync("SELECT * FROM musician;")
        if engine_is("postgres"):
            self.assertEqual(response, [{"id": 1, "name": ""}])

        if engine_is("cockroach"):
            self.assertEqual(
                response, [{"id": id[0]["id"], "name": ""}]  # type: ignore
            )

    @engines_only("postgres", "cockroach")
    def test_rename_table(self):
        """
        Test renaming a table with MigrationManager.
        """
        manager = MigrationManager()

        manager.rename_table(
            old_class_name="Manager",
            old_tablename="manager",
            new_class_name="Director",
            new_tablename="director",
        )

        asyncio.run(manager.run())

        if engine_is("postgres"):
            self.run_sync("INSERT INTO director VALUES (default, 'Dave');")

            response = self.run_sync("SELECT * FROM director;")
            self.assertEqual(response, [{"id": 1, "name": "Dave"}])

            # Reverse
            asyncio.run(manager.run(backwards=True))
            response = self.run_sync("SELECT * FROM manager;")
            self.assertEqual(response, [{"id": 1, "name": "Dave"}])

        if engine_is("cockroach"):
            id = 0
            id = self.run_sync(
                "INSERT INTO director VALUES (default, 'Dave') RETURNING id;"
            )

            response = self.run_sync("SELECT * FROM director;")
            self.assertEqual(response, [{"id": id[0]["id"], "name": "Dave"}])

            # Reverse
            asyncio.run(manager.run(backwards=True))
            response = self.run_sync("SELECT * FROM manager;")
            self.assertEqual(response, [{"id": id[0]["id"], "name": "Dave"}])

    @engines_only("postgres")
    def test_alter_column_unique(self):
        """
        Test altering a column uniqueness with MigrationManager.
        🐛 Cockroach bug: https://github.com/cockroachdb/cockroach/issues/42840 "unimplemented: cannot drop UNIQUE constraint "manager_name_key" using ALTER TABLE DROP CONSTRAINT, use DROP INDEX CASCADE instead"
        """  # noqa: E501
        manager = MigrationManager()

        manager.alter_column(
            table_class_name="Manager",
            tablename="manager",
            column_name="name",
            params={"unique": True},
            old_params={"unique": False},
        )

        asyncio.run(manager.run())

        with self.assertRaises(asyncpg.exceptions.UniqueViolationError):
            self.run_sync(
                "INSERT INTO manager VALUES "
                "(default, 'Dave'), (default, 'Dave');"
            )

        # Reverse
        asyncio.run(manager.run(backwards=True))
        self.run_sync(
            "INSERT INTO manager VALUES (default, 'Dave'), (default, 'Dave');"
        )
        response = self.run_sync("SELECT name FROM manager;")
        self.assertEqual(response, [{"name": "Dave"}, {"name": "Dave"}])

    @engines_only("postgres", "cockroach")
    def test_alter_column_set_null(self):
        """
        Test altering whether a column is nullable with MigrationManager.
        """
        manager = MigrationManager()

        manager.alter_column(
            table_class_name="Manager",
            tablename="manager",
            column_name="name",
            params={"null": True},
            old_params={"null": False},
        )

        asyncio.run(manager.run())
        self.assertTrue(
            self.get_postgres_is_nullable(
                tablename="manager", column_name="name"
            )
        )

        # Reverse
        asyncio.run(manager.run(backwards=True))
        self.assertFalse(
            self.get_postgres_is_nullable(
                tablename="manager", column_name="name"
            )
        )

    def _get_column_precision_and_scale(
        self, tablename="ticket", column_name="price"
    ):
        return self.run_sync(
            "SELECT numeric_precision, numeric_scale "
            "FROM information_schema.COLUMNS "
            f"WHERE table_name = '{tablename}' AND "
            f"column_name = '{column_name}';"
        )

    def _get_column_default(self, tablename="manager", column_name="name"):
        return self.run_sync(
            "SELECT column_default "
            "FROM information_schema.COLUMNS "
            f"WHERE table_name = '{tablename}' "
            f"AND column_name = '{column_name}';"
        )

    @engines_only("postgres")
    def test_alter_column_digits(self):
        """
        Test altering a column digits with MigrationManager.
        🐛 Cockroach bug: https://github.com/cockroachdb/cockroach/issues/49351 "ALTER COLUMN TYPE is not supported inside a transaction"
        """  # noqa: E501
        manager = MigrationManager()

        manager.alter_column(
            table_class_name="Ticket",
            tablename="ticket",
            column_name="price",
            params={"digits": (6, 2)},
            old_params={"digits": (5, 2)},
        )

        asyncio.run(manager.run())
        self.assertEqual(
            self._get_column_precision_and_scale(),
            [{"numeric_precision": 6, "numeric_scale": 2}],
        )

        asyncio.run(manager.run(backwards=True))
        self.assertEqual(
            self._get_column_precision_and_scale(),
            [{"numeric_precision": 5, "numeric_scale": 2}],
        )

    @engines_only("postgres")
    def test_alter_column_set_default(self):
        """
        Test altering a column default with MigrationManager.
        """
        manager = MigrationManager()

        manager.alter_column(
            table_class_name="Manager",
            tablename="manager",
            column_name="name",
            params={"default": "Unknown"},
            old_params={"default": ""},
        )

        asyncio.run(manager.run())
        self.assertEqual(
            self._get_column_default(),
            [{"column_default": "'Unknown'::character varying"}],
        )

        asyncio.run(manager.run(backwards=True))
        self.assertEqual(
            self._get_column_default(),
            [{"column_default": "''::character varying"}],
        )

    @engines_only("cockroach")
    def test_alter_column_set_default_alt(self):
        """
        Test altering a column default with MigrationManager.
        """
        manager = MigrationManager()

        manager.alter_column(
            table_class_name="Manager",
            tablename="manager",
            column_name="name",
            params={"default": "Unknown"},
            old_params={"default": ""},
        )

        asyncio.run(manager.run())
        self.assertEqual(
            self._get_column_default(),
            [{"column_default": "'Unknown':::STRING"}],
        )

        asyncio.run(manager.run(backwards=True))
        self.assertEqual(
            self._get_column_default(),
            [{"column_default": "'':::STRING"}],
        )

    @engines_only("postgres")
    def test_alter_column_drop_default(self):
        """
        Test setting a column default to None with MigrationManager.
        """
        # Make sure it has a non-null default to start with.
        manager_1 = MigrationManager()
        manager_1.alter_column(
            table_class_name="Manager",
            tablename="manager",
            column_name="name",
            params={"default": "Mr Manager"},
            old_params={"default": None},
        )
        asyncio.run(manager_1.run())
        self.assertEqual(
            self._get_column_default(),
            [{"column_default": "'Mr Manager'::character varying"}],
        )

        # Drop the default.
        manager_2 = MigrationManager()
        manager_2.alter_column(
            table_class_name="Manager",
            tablename="manager",
            column_name="name",
            params={"default": None},
            old_params={"default": "Mr Manager"},
        )
        asyncio.run(manager_2.run())
        self.assertEqual(
            self._get_column_default(),
            [{"column_default": None}],
        )

        # And add it back once more to be sure.
        manager_3 = manager_1
        asyncio.run(manager_3.run())
        self.assertEqual(
            self._get_column_default(),
            [{"column_default": "'Mr Manager'::character varying"}],
        )

        # Run them all backwards
        asyncio.run(manager_3.run(backwards=True))
        self.assertEqual(
            self._get_column_default(),
            [{"column_default": None}],
        )

        asyncio.run(manager_2.run(backwards=True))
        self.assertEqual(
            self._get_column_default(),
            [{"column_default": "'Mr Manager'::character varying"}],
        )

        asyncio.run(manager_1.run(backwards=True))
        self.assertEqual(
            self._get_column_default(),
            [{"column_default": None}],
        )

    @engines_only("cockroach")
    def test_alter_column_drop_default_alt(self):
        """
        Test setting a column default to None with MigrationManager.
        """
        # Make sure it has a non-null default to start with.
        manager_1 = MigrationManager()
        manager_1.alter_column(
            table_class_name="Manager",
            tablename="manager",
            column_name="name",
            params={"default": "Mr Manager"},
            old_params={"default": None},
        )
        asyncio.run(manager_1.run())
        self.assertEqual(
            self._get_column_default(),
            [{"column_default": "'Mr Manager':::STRING"}],
        )

        # Drop the default.
        manager_2 = MigrationManager()
        manager_2.alter_column(
            table_class_name="Manager",
            tablename="manager",
            column_name="name",
            params={"default": None},
            old_params={"default": "Mr Manager"},
        )
        asyncio.run(manager_2.run())
        self.assertEqual(
            self._get_column_default(),
            [{"column_default": None}],
        )

        # And add it back once more to be sure.
        manager_3 = manager_1
        asyncio.run(manager_3.run())
        self.assertEqual(
            self._get_column_default(),
            [{"column_default": "'Mr Manager':::STRING"}],
        )

        # Run them all backwards
        asyncio.run(manager_3.run(backwards=True))
        self.assertEqual(
            self._get_column_default(),
            [{"column_default": None}],
        )

        asyncio.run(manager_2.run(backwards=True))
        self.assertEqual(
            self._get_column_default(),
            [{"column_default": "'Mr Manager':::STRING"}],
        )

        asyncio.run(manager_1.run(backwards=True))
        self.assertEqual(
            self._get_column_default(),
            [{"column_default": None}],
        )

    @engines_only("postgres", "cockroach")
    def test_alter_column_add_index(self):
        """
        Test altering a column to add an index with MigrationManager.
        """
        manager = MigrationManager()

        manager.alter_column(
            table_class_name="Manager",
            tablename="manager",
            column_name="name",
            params={"index": True},
            old_params={"index": False},
        )

        asyncio.run(manager.run())
        self.assertTrue(
            Manager._get_index_name(["name"]) in Manager.indexes().run_sync()
        )

        asyncio.run(manager.run(backwards=True))
        self.assertTrue(
            Manager._get_index_name(["name"])
            not in Manager.indexes().run_sync()
        )

    @engines_only("postgres", "cockroach")
    def test_alter_column_set_type(self):
        """
        Test altering a column to change it's type with MigrationManager.
        """
        manager = MigrationManager()

        manager.alter_column(
            table_class_name="Manager",
            tablename="manager",
            column_name="name",
            params={},
            old_params={},
            column_class=Text,
            old_column_class=Varchar,
        )

        asyncio.run(manager.run())
        column_type_str = self.get_postgres_column_type(
            tablename="manager", column_name="name"
        )
        self.assertEqual(column_type_str, "TEXT")

        asyncio.run(manager.run(backwards=True))
        column_type_str = self.get_postgres_column_type(
            tablename="manager", column_name="name"
        )
        self.assertEqual(column_type_str, "CHARACTER VARYING")

    @engines_only("postgres")
    def test_alter_column_set_length(self):
        """
        Test altering a Varchar column's length with MigrationManager.
        🐛 Cockroach bug: https://github.com/cockroachdb/cockroach/issues/49351 "ALTER COLUMN TYPE is not supported inside a transaction"
        """  # noqa: E501
        manager = MigrationManager()

        manager.alter_column(
            table_class_name="Manager",
            tablename="manager",
            column_name="name",
            params={"length": 500},
            old_params={"length": 200},
            column_class=Text,
            old_column_class=Varchar,
        )

        asyncio.run(manager.run())
        self.assertEqual(
            self.get_postgres_varchar_length(
                tablename="manager", column_name="name"
            ),
            500,
        )

        asyncio.run(manager.run(backwards=True))
        self.assertEqual(
            self.get_postgres_varchar_length(
                tablename="manager", column_name="name"
            ),
            200,
        )

    @engines_only("postgres", "cockroach")
    @patch.object(
        BaseMigrationManager, "get_migration_managers", new_callable=AsyncMock
    )
    @patch.object(BaseMigrationManager, "get_app_config")
    def test_drop_table(
        self, get_app_config: MagicMock, get_migration_managers: MagicMock
    ):
        self.run_sync("DROP TABLE IF EXISTS musician;")

        name_column = Varchar()
        name_column._meta.name = "name"

        manager_1 = MigrationManager(migration_id="1", app_name="music")
        manager_1.add_table(
            class_name="Musician", tablename="musician", columns=[name_column]
        )
        asyncio.run(manager_1.run())

        manager_2 = MigrationManager(migration_id="2", app_name="music")
        manager_2.drop_table(class_name="Musician", tablename="musician")
        asyncio.run(manager_2.run())

        self.assertTrue(not self.table_exists("musician"))

        # Reverse
        get_migration_managers.return_value = [manager_1]
        app_config = AppConfig(app_name="music", migrations_folder_path="")
        get_app_config.return_value = app_config
        asyncio.run(manager_2.run(backwards=True))

        get_migration_managers.assert_called_with(
            app_config=app_config, max_migration_id="2", offset=-1
        )
        self.assertTrue(self.table_exists("musician"))

        self.run_sync("DROP TABLE IF EXISTS musician;")

    @engines_only("postgres", "cockroach")
    def test_change_table_schema(self):
        manager = MigrationManager(migration_id="1", app_name="music")

        manager.change_table_schema(
            class_name="Manager",
            tablename="manager",
            new_schema="schema_1",
            old_schema=None,
        )

        # Preview
        manager.preview = True
        with patch("sys.stdout", new=StringIO()) as fake_out:
            asyncio.run(manager.run())

            output = fake_out.getvalue()

            self.assertEqual(
                output,
                '  - 1 [preview forwards]... CREATE SCHEMA IF NOT EXISTS "schema_1"\nALTER TABLE "manager" SET SCHEMA "schema_1"\n',  # noqa: E501
            )



import time
from enum import Enum
from unittest import TestCase

import pytest

from piccolo.columns.column_types import JSON, JSONB, Array, Integer, Varchar
from piccolo.table import Table
from piccolo.utils.sql_values import convert_to_sql_value


class TestConvertToSQLValue(TestCase):
    def test_convert_json(self):
        """
        Make sure Python objects are serialised correctly to JSON strings.
        """
        self.assertEqual(
            convert_to_sql_value(value={"a": 123}, column=JSON()).replace(
                " ", ""
            ),
            '{"a":123}',
        )

        self.assertEqual(
            convert_to_sql_value(value={"a": 123}, column=JSONB()).replace(
                " ", ""
            ),
            '{"a":123}',
        )

    def test_convert_table_instance(self):
        """
        Make sure Table instances are converted to integers.
        """

        class MyTable(Table):
            pass

        instance = MyTable(id=1)

        self.assertEqual(
            convert_to_sql_value(
                value=instance, column=MyTable._meta.primary_key
            ),
            1,
        )

    def test_convert_enum(self):
        """
        Make sure Enum instances are converted to their values.
        """

        class Colour(Enum):
            red = "r"

        self.assertEqual(
            convert_to_sql_value(value=Colour.red, column=Varchar()), "r"
        )

    def test_other(self):
        """
        Make sure simple Python values are returned correctly.
        """
        self.assertEqual(
            convert_to_sql_value(value=1, column=Integer()),
            1,
        )

    def test_convert_enum_list(self):
        """
        It's possible to have a list of enums when using ``Array`` columns.
        """

        class Colour(Enum):
            red = "r"
            green = "g"
            blue = "b"

        self.assertEqual(
            convert_to_sql_value(
                value=[Colour.red, Colour.green, Colour.blue],
                column=Array(Varchar()),
            ),
            ["r", "g", "b"],
        )

    @pytest.mark.speed
    def test_convert_large_list(self):
        """
        Large lists are problematic. We need to check each value in the list,
        but as efficiently as possible.
        """
        start = time.time()

        convert_to_sql_value(
            value=[i for i in range(1000)],
            column=Array(Varchar()),
        )

        duration = time.time() - start
        print(duration)



from unittest import TestCase

from piccolo.columns.column_types import JSON
from piccolo.table import Table


class MyTable(Table):
    json = JSON()


class MyTableDefault(Table):
    """
    Test the different default types.
    """

    json = JSON()
    json_str = JSON(default="{}")
    json_dict = JSON(default={})
    json_list = JSON(default=[])
    json_none = JSON(default=None, null=True)


class TestJSONSave(TestCase):
    def setUp(self):
        MyTable.create_table().run_sync()

    def tearDown(self):
        MyTable.alter().drop_table().run_sync()

    def test_json_string(self):
        """
        Test storing a valid JSON string.
        """
        row = MyTable(json='{"a": 1}')
        row.save().run_sync()

        row_from_db = MyTable.select(MyTable.json).first().run_sync()
        assert row_from_db is not None

        self.assertEqual(
            row_from_db["json"].replace(" ", ""),
            '{"a":1}',
        )

    def test_json_object(self):
        """
        Test storing a valid JSON object.
        """
        row = MyTable(json={"a": 1})
        row.save().run_sync()

        row_from_db = MyTable.select(MyTable.json).first().run_sync()
        assert row_from_db is not None

        self.assertEqual(
            row_from_db["json"].replace(" ", ""),
            '{"a":1}',
        )


class TestJSONDefault(TestCase):
    def setUp(self):
        MyTableDefault.create_table().run_sync()

    def tearDown(self):
        MyTableDefault.alter().drop_table().run_sync()

    def test_json_default(self):
        row = MyTableDefault()
        row.save().run_sync()

        self.assertEqual(row.json, "{}")
        self.assertEqual(row.json_str, "{}")
        self.assertEqual(row.json_dict, "{}")
        self.assertEqual(row.json_list, "[]")
        self.assertEqual(row.json_none, None)

    def test_invalid_default(self):
        with self.assertRaises(ValueError):
            for value in ("a", 1, ("x", "y", "z")):
                JSON(default=value)  # type: ignore


class TestJSONInsert(TestCase):
    def setUp(self):
        MyTable.create_table().run_sync()

    def tearDown(self):
        MyTable.alter().drop_table().run_sync()

    def check_response(self):
        row = MyTable.select(MyTable.json).first().run_sync()
        assert row is not None
        self.assertEqual(
            row["json"].replace(" ", ""),
            '{"message":"original"}',
        )

    def test_json_string(self):
        """
        Test inserting using a string.
        """
        row = MyTable(json='{"message": "original"}')
        MyTable.insert(row).run_sync()
        self.check_response()

    def test_json_object(self):
        """
        Test inserting using an object.
        """
        row = MyTable(json={"message": "original"})
        MyTable.insert(row).run_sync()


class TestJSONUpdate(TestCase):
    def setUp(self):
        MyTable.create_table().run_sync()

    def tearDown(self):
        MyTable.alter().drop_table().run_sync()

    def add_row(self):
        row = MyTable(json={"message": "original"})
        row.save().run_sync()

    def check_response(self):
        row = MyTable.select(MyTable.json).first().run_sync()
        assert row is not None
        self.assertEqual(
            row["json"].replace(" ", ""),
            '{"message":"updated"}',
        )

    def test_json_update_string(self):
        """
        Test updating a JSON field using a string.
        """
        self.add_row()
        MyTable.update(
            {MyTable.json: '{"message": "updated"}'}, force=True
        ).run_sync()
        self.check_response()

    def test_json_update_object(self):
        """
        Test updating a JSON field using an object.
        """
        self.add_row()
        MyTable.update(
            {MyTable.json: {"message": "updated"}}, force=True
        ).run_sync()
        self.check_response()



from __future__ import annotations

import typing as t
from abc import ABCMeta, abstractmethod
from dataclasses import dataclass
from datetime import datetime
from importlib.util import find_spec
from string import Formatter

if t.TYPE_CHECKING:  # pragma: no cover
    from piccolo.table import Table
    from piccolo.columns import Column

from uuid import UUID

if find_spec("asyncpg"):
    from asyncpg.pgproto.pgproto import UUID as apgUUID
else:
    apgUUID = UUID


class Selectable(metaclass=ABCMeta):
    """
    Anything which inherits from this can be used in a select query.
    """

    __slots__ = ("_alias",)

    _alias: t.Optional[str]

    @abstractmethod
    def get_select_string(
        self, engine_type: str, with_alias: bool = True
    ) -> QueryString:
        """
        In a query, what to output after the select statement - could be a
        column name, a sub query, a function etc. For a column it will be the
        column name.
        """
        raise NotImplementedError()

    def as_alias(self, alias: str) -> Selectable:
        """
        Allows column names to be changed in the result of a select.
        """
        self._alias = alias
        return self


@dataclass
class Fragment:
    prefix: str
    index: int = 0
    no_arg: bool = False


class QueryString(Selectable):
    """
    When we're composing complex queries, we're combining QueryStrings, rather
    than concatenating strings directly. The reason for this is QueryStrings
    keep the parameters separate, so we can pass parameterised queries to the
    engine - which helps prevent SQL Injection attacks.
    """

    __slots__ = (
        "template",
        "args",
        "query_type",
        "table",
        "_frozen_compiled_strings",
        "columns",
    )

    def __init__(
        self,
        template: str,
        *args: t.Any,
        query_type: str = "generic",
        table: t.Optional[t.Type[Table]] = None,
        alias: t.Optional[str] = None,
    ) -> None:
        """
        :param template:
            The SQL query, with curly brackets as placeholders for any values::

                "WHERE {} = {}"

        :param args:
            The values to insert (one value is needed for each set of curly
            braces in the template).
        :param query_type:
            The query type is sometimes used by the engine to modify how the
            query is run. For example, INSERT queries on old SQLite versions.
        :param table:
            Sometimes the ``piccolo.engine.base.Engine`` needs access to the
            table that the query is being run on.

        """
        self.template = template
        self.query_type = query_type
        self.table = table
        self._frozen_compiled_strings: t.Optional[
            t.Tuple[str, t.List[t.Any]]
        ] = None
        self._alias = alias
        self.args, self.columns = self.process_args(args)

    def process_args(
        self, args: t.Sequence[t.Any]
    ) -> t.Tuple[t.Sequence[t.Any], t.Sequence[Column]]:
        """
        If a Column is passed in, we convert it to the name of the column
        (including joins).
        """
        from piccolo.columns import Column

        processed_args = []
        columns = []

        for arg in args:
            if isinstance(arg, Column):
                columns.append(arg)
                arg = QueryString(
                    f"{arg._meta.get_full_name(with_alias=False)}"
                )
            elif isinstance(arg, QueryString):
                columns.extend(arg.columns)

            processed_args.append(arg)

        return (processed_args, columns)

    def as_alias(self, alias: str) -> QueryString:
        self._alias = alias
        return self

    def __str__(self):
        """
        The SQL returned by the ``__str__`` method isn't used directly in
        queries - it's just a usability feature.
        """
        _, bundled, combined_args = self.bundle(
            start_index=1, bundled=[], combined_args=[]
        )
        template = "".join(
            fragment.prefix + ("" if fragment.no_arg else "{}")
            for fragment in bundled
        )

        # Do some basic type conversion here.
        converted_args = []
        for arg in combined_args:
            _type = type(arg)
            if _type == str:
                converted_args.append(f"'{arg}'")
            elif _type == datetime:
                dt_string = arg.isoformat()
                converted_args.append(f"'{dt_string}'")
            elif _type == UUID or _type == apgUUID:
                converted_args.append(f"'{arg}'")
            elif arg is None:
                converted_args.append("null")
            else:
                converted_args.append(arg)

        return template.format(*converted_args)

    def bundle(
        self,
        start_index: int = 1,
        bundled: t.Optional[t.List[Fragment]] = None,
        combined_args: t.Optional[t.List] = None,
    ):
        # Split up the string, separating by {}.
        fragments = [
            Fragment(prefix=i[0]) for i in Formatter().parse(self.template)
        ]

        bundled = [] if bundled is None else bundled
        combined_args = [] if combined_args is None else combined_args

        for index, fragment in enumerate(fragments):
            try:
                value = self.args[index]
            except IndexError:
                # trailing element
                fragment.no_arg = True
                bundled.append(fragment)
            else:
                if isinstance(value, QueryString):
                    fragment.no_arg = True
                    bundled.append(fragment)

                    start_index, _, _ = value.bundle(
                        start_index=start_index,
                        bundled=bundled,
                        combined_args=combined_args,
                    )
                else:
                    fragment.index = start_index
                    bundled.append(fragment)
                    start_index += 1
                    combined_args.append(value)

        return (start_index, bundled, combined_args)

    def compile_string(
        self, engine_type: str = "postgres"
    ) -> t.Tuple[str, t.List[t.Any]]:
        """
        Compiles the template ready for the engine - keeping the arguments
        separate from the template.
        """
        if self._frozen_compiled_strings is not None:
            return self._frozen_compiled_strings

        _, bundled, combined_args = self.bundle(
            start_index=1, bundled=[], combined_args=[]
        )
        if engine_type in ("postgres", "cockroach"):
            string = "".join(
                fragment.prefix
                + ("" if fragment.no_arg else f"${fragment.index}")
                for fragment in bundled
            )

        elif engine_type == "sqlite":
            string = "".join(
                fragment.prefix + ("" if fragment.no_arg else "?")
                for fragment in bundled
            )

        else:
            raise Exception("Engine type not recognised")

        return (string, combined_args)

    def freeze(self, engine_type: str = "postgres"):
        self._frozen_compiled_strings = self.compile_string(
            engine_type=engine_type
        )

    ###########################################################################

    def get_select_string(
        self, engine_type: str, with_alias: bool = True
    ) -> QueryString:
        if with_alias and self._alias:
            return QueryString("{} AS " + self._alias, self)
        else:
            return self

    def get_where_string(self, engine_type: str) -> QueryString:
        return self.get_select_string(
            engine_type=engine_type, with_alias=False
        )

    ###########################################################################
    # Basic logic

    def __eq__(self, value) -> QueryString:  # type: ignore[override]
        return QueryString("{} = {}", self, value)

    def __ne__(self, value) -> QueryString:  # type: ignore[override]
        return QueryString("{} != {}", self, value)

    def __add__(self, value) -> QueryString:
        return QueryString("{} + {}", self, value)

    def __sub__(self, value) -> QueryString:
        return QueryString("{} - {}", self, value)

    def is_in(self, value) -> QueryString:
        return QueryString("{} IN {}", self, value)

    def not_in(self, value) -> QueryString:
        return QueryString("{} NOT IN {}", self, value)


class Unquoted(QueryString):
    """
    This is deprecated - just use QueryString directly.
    """

    pass



from __future__ import annotations

import typing as t

from piccolo.custom_types import Combinable
from piccolo.query.base import Query
from piccolo.query.functions.aggregate import Count as CountFunction
from piccolo.query.mixins import WhereDelegate
from piccolo.querystring import QueryString

if t.TYPE_CHECKING:  # pragma: no cover
    from piccolo.columns import Column
    from piccolo.table import Table


class Count(Query):

    __slots__ = ("where_delegate", "column", "_distinct")

    def __init__(
        self,
        table: t.Type[Table],
        column: t.Optional[Column] = None,
        distinct: t.Optional[t.Sequence[Column]] = None,
        **kwargs,
    ):
        super().__init__(table, **kwargs)
        self.column = column
        self._distinct = distinct
        self.where_delegate = WhereDelegate()

    ###########################################################################
    # Clauses

    def where(self: Self, *where: t.Union[Combinable, QueryString]) -> Self:
        self.where_delegate.where(*where)
        return self

    def distinct(self: Self, columns: t.Optional[t.Sequence[Column]]) -> Self:
        self._distinct = columns
        return self

    ###########################################################################

    async def response_handler(self, response) -> bool:
        return response[0]["count"]

    @property
    def default_querystrings(self) -> t.Sequence[QueryString]:
        table: t.Type[Table] = self.table

        query = table.select(
            CountFunction(column=self.column, distinct=self._distinct)
        )

        query.where_delegate._where = self.where_delegate._where

        return query.querystrings


Self = t.TypeVar("Self", bound=Count)
